#    Licensed under the Apache License, Version 2.0 (the "License"); you may
#    not use this file except in compliance with the License. You may obtain
#    a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#    WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#    License for the specific language governing permissions and limitations
#    under the License.

import flask

from sushy_tools.emulator import api_utils
from sushy_tools import error


virtual_media = flask.Blueprint(
    'VirtualMedia', __name__,
    url_prefix='/redfish/v1/Managers/<identity>/VirtualMedia')


@virtual_media.route('', methods=['GET'])
@api_utils.returns_json
def virtual_media_collection_resource(identity):
    api_utils.debug('Serving virtual media resources for manager "%s"',
                    identity)

    return flask.render_template(
        'virtual_media_collection.json',
        identity=identity,
        uuid=flask.current_app.managers.get_manager(identity)['UUID'],
        devices=flask.current_app.vmedia.devices
    )


@virtual_media.route('/<device>', methods=['GET'])
@api_utils.returns_json
def virtual_media_resource(identity, device):
    device_name = flask.current_app.vmedia.get_device_name(
        identity, device)

    media_types = flask.current_app.vmedia.get_device_media_types(
        identity, device)

    device_info = flask.current_app.vmedia.get_device_image_info(
        identity, device)

    api_utils.debug('Serving virtual media %s at manager "%s"',
                    device, identity)

    return flask.render_template(
        'virtual_media.json',
        identity=identity,
        device=device,
        name=device_name,
        media_types=media_types,
        image_url=device_info.image_url,
        image_name=device_info.image_name,
        inserted=device_info.inserted,
        write_protected=device_info.write_protected,
        username=device_info.username,
        password=device_info.password,
        verify_certificate=device_info.verify,
    )


@virtual_media.route('/<device>', methods=['PATCH'])
@api_utils.returns_json
def virtual_media_patch(identity, device):
    flask.current_app.managers.get_manager(identity)
    if not flask.request.json:
        raise error.BadRequest("Empty or malformed patch")

    api_utils.debug('Updating virtual media %s at manager "%s"',
                    device, identity)

    verify = flask.request.json.get('VerifyCertificate')
    if verify is not None:
        if not isinstance(verify, bool):
            raise error.BadRequest("VerifyCertificate must be a boolean")

        flask.current_app.vmedia.update_device_info(
            identity, device, verify=verify)
        return '', 204
    else:
        raise error.BadRequest("Empty or malformed patch")


@virtual_media.route('/<device>/Certificates', methods=['GET'])
@api_utils.returns_json
def virtual_media_certificates(identity, device):
    flask.current_app.managers.get_manager(identity)
    location = \
        f'/redfish/v1/Managers/{identity}/VirtualMedia/{device}/Certificates'
    certificates = flask.current_app.vmedia.list_certificates(identity, device)
    return flask.render_template(
        'certificate_collection.json',
        location=location,
        certificates=[cert.id for cert in certificates],
    )


@virtual_media.route('/<device>/Certificates', methods=['POST'])
@api_utils.returns_json
def virtual_media_add_certificate(identity, device):
    flask.current_app.managers.get_manager(identity)
    if not flask.request.json:
        raise error.BadRequest("Empty or malformed certificate")

    try:
        cert_string = flask.request.json['CertificateString']
        cert_type = flask.request.json['CertificateType']
    except KeyError as exc:
        raise error.BadRequest(f"Missing required parameter {exc}")

    if cert_type != 'PEM':
        raise error.BadRequest(
            f"Only PEM certificates are supported, got {cert_type}")

    api_utils.debug('Adding certificate for virtual media %s at '
                    'manager "%s"', device, identity)

    cert = flask.current_app.vmedia.add_certificate(
        identity, device, cert_string, cert_type)

    location = (
        f'/redfish/v1/Managers/{identity}/VirtualMedia/{device}'
        f'/Certificates/{cert.id}'
    )
    return '', 204, {'Location': location}


@virtual_media.route('/<device>/Certificates/<cert_id>', methods=['GET'])
@api_utils.returns_json
def virtual_media_get_certificate(identity, device, cert_id):
    flask.current_app.managers.get_manager(identity)
    location = (
        f'/redfish/v1/Managers/{identity}/VirtualMedia/{device}'
        f'/Certificates/{cert_id}'
    )
    certificates = flask.current_app.vmedia.list_certificates(identity, device)
    try:
        cert = next(c for c in certificates if c.id == cert_id)
    except StopIteration:
        raise error.NotFound()

    return flask.render_template(
        'certificate.json',
        location=location,
        cert_id=cert.id,
        cert_string=cert.string,
        cert_type=cert.type_,
    )


@virtual_media.route('/<device>/Certificates/<cert_id>', methods=['DELETE'])
@api_utils.returns_json
def virtual_media_delete_certificate(identity, device, cert_id):
    flask.current_app.managers.get_manager(identity)
    api_utils.debug('Removing certificate %s for virtual media %s at '
                    'manager "%s"', cert_id, device, identity)
    flask.current_app.vmedia.delete_certificate(identity, device, cert_id)
    return '', 204


@virtual_media.route('/<device>/Actions/VirtualMedia.InsertMedia',
                     methods=['POST'])
@api_utils.returns_json
def virtual_media_insert(identity, device):
    image = flask.request.json.get('Image')
    inserted = flask.request.json.get('Inserted', True)
    write_protected = flask.request.json.get('WriteProtected', True)
    username = flask.request.json.get('UserName', '')
    password = flask.request.json.get('Password', '')

    if (not username and password) or (username and not password):
        message = "UserName and Password must be passed together"
        return flask.render_template('error.json', message=message), 400

    manager = flask.current_app.managers.get_manager(identity)
    systems = flask.current_app.managers.get_managed_systems(manager)
    if not systems:
        api_utils.warning('Manager %s manages no systems', identity)
        return '', 204

    image_path = flask.current_app.vmedia.insert_image(
        identity, device, image, inserted, write_protected,
        username=username, password=password)

    for system in systems:
        try:
            flask.current_app.systems.set_boot_image(
                system, device, boot_image=image_path,
                write_protected=write_protected)

        except error.NotSupportedError as ex:
            api_utils.warning(
                'System %s failed to set boot image %s on device %s: '
                '%s', system, image_path, device, ex)

    api_utils.info(
        'Virtual media placed into device %(dev)s of manager %(mgr)s for '
        'systems %(sys)s. Image %(img)s inserted %(ins)s',
        {'dev': device, 'mgr': identity, 'sys': systems,
         'img': image or '<empty>', 'ins': inserted})

    return '', 204


@virtual_media.route('/<device>/Actions/VirtualMedia.EjectMedia',
                     methods=['POST'])
@api_utils.returns_json
def virtual_media_eject(identity, device):
    flask.current_app.vmedia.eject_image(identity, device)

    manager = flask.current_app.managers.get_manager(identity)
    systems = flask.current_app.managers.get_managed_systems(manager)
    if not systems:
        api_utils.warning('Manager %s manages no systems', identity)
        return '', 204

    for system in systems:
        try:
            flask.current_app.systems.set_boot_image(system, device)

        except error.NotSupportedError as ex:
            api_utils.warning(
                'System %s failed to remove boot image from device %s: '
                '%s', system, device, ex)

    api_utils.info(
        'Virtual media ejected from device %s manager %s systems %s',
        device, identity, systems)

    return '', 204



#    Licensed under the Apache License, Version 2.0 (the "License"); you may
#    not use this file except in compliance with the License. You may obtain
#    a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#    WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#    License for the specific language governing permissions and limitations
#    under the License.

import time
from unittest import mock

from oslotest import base

from sushy_tools.emulator.resources.systems import fakedriver
from sushy_tools import error


UUID = fakedriver.DEFAULT_UUID


class FakeDriverTestCase(base.BaseTestCase):

    def setUp(self):
        super().setUp()
        test_driver_class = fakedriver.FakeDriver.initialize(
            {}, mock.MagicMock())
        self.cache = {}
        with mock.patch('sushy_tools.emulator.memoize.PersistentDict',
                        return_value=self.cache, autospec=True):
            self.test_driver = test_driver_class()

    def test_systems(self):
        self.assertEqual([UUID], self.test_driver.systems)
        self.assertEqual('fake', self.test_driver.name(UUID))
        self.assertEqual(UUID, self.test_driver.uuid('fake'))
        self.assertEqual(UUID, self.test_driver.uuid(UUID))
        self.assertEqual('fake', self.test_driver.name('fake'))
        self.assertRaises(error.NotFound, self.test_driver.uuid, 'foo')
        self.assertRaises(error.NotFound, self.test_driver.name, 'foo')

    @mock.patch('random.randint', autospec=True, return_value=0)
    def test_power_state(self, mock_rand):
        self.assertEqual('Off', self.test_driver.get_power_state(UUID))
        self.test_driver.set_power_state(UUID, 'On')
        self.assertEqual('On', self.test_driver.get_power_state(UUID))
        self.test_driver.set_power_state(UUID, 'ForceOff')
        self.assertEqual('Off', self.test_driver.get_power_state(UUID))

    @mock.patch('random.randint', autospec=True, return_value=1000)
    def test_power_state_delay(self, mock_rand):
        self.assertEqual('Off', self.test_driver.get_power_state(UUID))
        self.test_driver.set_power_state(UUID, 'On')
        self.assertEqual('Off', self.test_driver.get_power_state(UUID))

        new_time = time.time() + 2000
        with mock.patch.object(time, 'time', autospec=True,
                               return_value=new_time):
            self.assertEqual('On', self.test_driver.get_power_state(UUID))

    @mock.patch('random.randint', autospec=True, return_value=1000)
    def test_reboot_delay(self, mock_rand):
        self.cache[UUID]['power_state'] = 'On'

        self.assertEqual('On', self.test_driver.get_power_state(UUID))
        self.test_driver.set_power_state(UUID, 'ForceRestart')
        self.assertEqual('Off', self.test_driver.get_power_state(UUID))

        new_time = time.time() + 2000
        with mock.patch.object(time, 'time', autospec=True,
                               return_value=new_time):
            self.assertEqual('On', self.test_driver.get_power_state(UUID))

    def test_boot_mode(self):
        self.assertEqual('UEFI', self.test_driver.get_boot_mode(UUID))
        self.test_driver.set_boot_mode(UUID, 'legacy')
        self.assertEqual('legacy', self.test_driver.get_boot_mode(UUID))

    def test_boot_device(self):
        self.assertEqual('Hdd', self.test_driver.get_boot_device(UUID))
        self.test_driver.set_boot_device(UUID, 'Cd')
        self.assertEqual('Cd', self.test_driver.get_boot_device(UUID))

    def test_boot_image(self):
        self.assertEqual((None, False, False),
                         self.test_driver.get_boot_image(UUID, 'Cd'))
        self.test_driver.set_boot_image(UUID, 'Cd', 'http://example')
        self.assertEqual(('http://example', True, True),
                         self.test_driver.get_boot_image(UUID, 'Cd'))
        self.assertEqual((None, False, False),
                         self.test_driver.get_boot_image(UUID, 'Hdd'))



# Copyright 2018 Red Hat, Inc.
# All Rights Reserved.
#
#    Licensed under the Apache License, Version 2.0 (the "License"); you may
#    not use this file except in compliance with the License. You may obtain
#    a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#    WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#    License for the specific language governing permissions and limitations
#    under the License.

import pickle
import sqlite3
from unittest import mock

from oslotest import base

from sushy_tools.emulator import memoize


class MemoizeTestCase(base.BaseTestCase):

    def test_local_cache(self):

        class Driver(object):
            call_count = 0

            @memoize.memoize()
            def fun(self, *args, **kwargs):
                self.call_count += 1
                return args, kwargs

        driver = Driver()

        result = driver.fun(1, x=2)
        self.assertEqual(((1,), {'x': 2}), result)

        result = driver.fun(1, x=2)
        self.assertEqual(((1,), {'x': 2}), result)

        # Due to memoization, expect just one call
        self.assertEqual(1, driver.call_count)

        driver = Driver()

        result = driver.fun(1, x=2)
        self.assertEqual(((1,), {'x': 2}), result)

        # Due to volatile cache, expect one more call
        self.assertEqual(1, driver.call_count)

    def test_external_cache(self):

        permanent_cache = {}

        class Driver(object):
            call_count = 0

            @memoize.memoize(permanent_cache=permanent_cache)
            def fun(self, *args, **kwargs):
                self.call_count += 1
                return args, kwargs

        driver = Driver()

        result = driver.fun(1, x=2)
        self.assertEqual(((1,), {'x': 2}), result)

        result = driver.fun(1, x=2)
        self.assertEqual(((1,), {'x': 2}), result)

        # Due to memoization, expect just one call
        self.assertEqual(1, driver.call_count)

        driver = Driver()

        result = driver.fun(1, x=2)
        self.assertEqual(((1,), {'x': 2}), result)

        # Due to permanent cache, expect no more calls
        self.assertEqual(0, driver.call_count)


@mock.patch.object(sqlite3, 'connect', autospec=True)
class PersistentDictTestCase(base.BaseTestCase):

    def test_make_permanent(self, mock_sqlite3):
        pd = memoize.PersistentDict()
        pd.make_permanent('/', 'file')
        mock_sqlite3.assert_called_once_with('/file.sqlite')

    def test_encode(self, mock_sqlite3):
        pd = memoize.PersistentDict()
        value = pd.encode({1: '2'})
        self.assertEqual(pickle.dumps({1: '2'}), value)

    def test_decode(self, mock_sqlite3):
        pd = memoize.PersistentDict()
        value = pd.decode(pickle.dumps({1: '2'}))
        self.assertEqual({1: '2'}, value)

    def test___getitem__(self, mock_sqlite3):
        pd = memoize.PersistentDict()
        pd.make_permanent('/', 'file')

        mock_conn = mock_sqlite3.return_value.__enter__.return_value
        mock_cursor = mock_conn.cursor.return_value
        mock_cursor.execute.reset_mock()
        mock_cursor.fetchone.return_value = [pickle.dumps('pickled-value')]

        result = pd[1]
        self.assertEqual('pickled-value', result)

        mock_cursor.execute.assert_called_with(
            'select value from cache where key=?', (pickle.dumps(1),))

    def test___getitem__retries(self, mock_sqlite3):
        pd = memoize.PersistentDict()
        pd.make_permanent('/', 'file')

        mock_conn = mock_sqlite3.return_value.__enter__.return_value
        mock_cursor = mock_conn.cursor.return_value
        mock_cursor.execute.reset_mock()
        mock_cursor.execute.side_effect = [
            sqlite3.OperationalError,
            None
        ]
        mock_cursor.fetchone.return_value = [pickle.dumps('pickled-value')]

        result = pd[1]
        self.assertEqual('pickled-value', result)

        mock_cursor.execute.assert_called_with(
            'select value from cache where key=?', (pickle.dumps(1),))
        self.assertEqual(2, mock_cursor.execute.call_count)

    def test___setitem__(self, mock_sqlite3):
        pd = memoize.PersistentDict()
        pd.make_permanent('/', 'file')

        mock_conn = mock_sqlite3.return_value.__enter__.return_value
        mock_cursor = mock_conn.cursor.return_value
        mock_cursor.execute.reset_mock()

        pd[1] = 2

        mock_cursor.execute.assert_called_once_with(
            'insert or replace into cache values (?, ?)',
            (pickle.dumps(1), pickle.dumps(2)))

    def test___setitem__retries(self, mock_sqlite3):
        pd = memoize.PersistentDict()
        pd.make_permanent('/', 'file')

        mock_conn = mock_sqlite3.return_value.__enter__.return_value
        mock_cursor = mock_conn.cursor.return_value
        mock_cursor.execute.reset_mock()
        mock_cursor.execute.side_effect = [
            sqlite3.OperationalError,
            None
        ]

        pd[1] = 2

        mock_cursor.execute.assert_called_with(
            'insert or replace into cache values (?, ?)',
            (pickle.dumps(1), pickle.dumps(2)))
        self.assertEqual(2, mock_cursor.execute.call_count)

    def test___delitem__(self, mock_sqlite3):
        pd = memoize.PersistentDict()
        pd.make_permanent('/', 'file')

        mock_conn = mock_sqlite3.return_value.__enter__.return_value
        mock_cursor = mock_conn.cursor.return_value
        mock_cursor.execute.reset_mock()

        del pd[1]

        mock_cursor.execute.assert_called_once_with(
            'delete from cache where key=?', (pickle.dumps(1),))

    def test___delitem__fails(self, mock_sqlite3):
        pd = memoize.PersistentDict()
        pd.make_permanent('/', 'file')

        mock_conn = mock_sqlite3.return_value.__enter__.return_value
        mock_cursor = mock_conn.cursor.return_value
        mock_cursor.execute.reset_mock()
        mock_cursor.rowcount = 0

        def _del():
            del pd[1]

        self.assertRaises(KeyError, _del)

        mock_cursor.execute.assert_called_once_with(
            'delete from cache where key=?', (pickle.dumps(1),))

    def test___iter__(self, mock_sqlite3):
        pd = memoize.PersistentDict()
        pd.make_permanent('/', 'file')

        mock_conn = mock_sqlite3.return_value.__enter__.return_value
        mock_cursor = mock_conn.cursor.return_value
        mock_cursor.execute.reset_mock()
        mock_cursor.fetchall.return_value = [[pickle.dumps('pickled-key')]]

        self.assertEqual(['pickled-key'], list(iter(pd)))

        mock_cursor.execute.assert_called_once_with('select key from cache')

    def test___len__(self, mock_sqlite3):
        pd = memoize.PersistentDict()
        pd.make_permanent('/', 'file')

        mock_conn = mock_sqlite3.return_value.__enter__.return_value
        mock_cursor = mock_conn.cursor.return_value
        mock_cursor.execute.reset_mock()
        mock_cursor.fetchone.return_value = [42]

        self.assertEqual(42, len(pd))

        mock_cursor.execute.assert_called_once_with(
            'select count(*) from cache')



# Copyright 2019 Red Hat, Inc.
# All Rights Reserved.
#
#    Licensed under the Apache License, Version 2.0 (the "License"); you may
#    not use this file except in compliance with the License. You may obtain
#    a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#    WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#    License for the specific language governing permissions and limitations
#    under the License.
from unittest import mock

from oslotest import base

from sushy_tools.emulator.resources.volumes import StaticDriver


@mock.patch('sushy_tools.emulator.resources.volumes.memoize.PersistentDict',
            new=dict)
class StaticDriverTestCase(base.BaseTestCase):

    SYSTEM_UUID = "da69abcc-dae0-4913-9a7b-d344043097c0"
    STORAGE_ID = "1"
    VOLUMES_COL = [
        {
            "libvirtPoolName": "sushyPool",
            "libvirtVolName": "testVol",
            "Id": "1",
            "Name": "Sample Volume 1",
            "VolumeType": "Mirrored",
            "CapacityBytes": 23748
        },
        {
            "libvirtPoolName": "sushyPool",
            "libvirtVolName": "testVol1",
            "Id": "2",
            "Name": "Sample Volume 2",
            "VolumeType": "StripedWithParity",
            "CapacityBytes": 48395
        }
    ]

    CONFIG = {
        'SUSHY_EMULATOR_VOLUMES': {
            (SYSTEM_UUID, STORAGE_ID): VOLUMES_COL
        }
    }

    def setUp(self):
        super().setUp()
        self.test_driver = StaticDriver(self.CONFIG, mock.MagicMock())

    def test_get_volumes_col(self):
        vol_col = self.test_driver.get_volumes_col(self.SYSTEM_UUID,
                                                   self.STORAGE_ID)
        self.assertEqual(self.VOLUMES_COL, vol_col)

    def test_add_volume(self):
        vol = {
            "libvirtPoolName": "sushyPool",
            "libvirtVolName": "testVol2",
            "Id": "3",
            "Name": "Sample Volume 3",
            "VolumeType": "Mirrored",
            "CapacityBytes": 76584
        }
        self.test_driver.add_volume(self.SYSTEM_UUID, self.STORAGE_ID, vol)
        vol_col = self.test_driver.get_volumes_col(self.SYSTEM_UUID,
                                                   self.STORAGE_ID)
        self.assertIn(vol, vol_col)

    def test_delete_volume(self):
        vol = {
            "libvirtPoolName": "sushyPool",
            "libvirtVolName": "testVol",
            "Id": "1",
            "Name": "Sample Volume 1",
            "VolumeType": "Mirrored",
            "CapacityBytes": 23748
        }
        self.test_driver.delete_volume(self.SYSTEM_UUID, self.STORAGE_ID, vol)
        vol_col = self.test_driver.get_volumes_col(self.SYSTEM_UUID,
                                                   self.STORAGE_ID)
        self.assertNotIn(vol, vol_col)



# Copyright 2019 Red Hat, Inc.
# All Rights Reserved.
#
#    Licensed under the Apache License, Version 2.0 (the "License"); you may
#    not use this file except in compliance with the License. You may obtain
#    a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#    WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#    License for the specific language governing permissions and limitations
#    under the License.

import builtins
from unittest import mock

from oslotest import base

from sushy_tools.emulator.resources import vmedia
from sushy_tools import error


class StaticDriverTestCase(base.BaseTestCase):

    UUID = 'ZZZ-YYY-XXX'

    CONFIG = {
        'SUSHY_EMULATOR_VMEDIA_DEVICES': {
            "Cd": {
                "Name": "Virtual CD",
                "MediaTypes": [
                    "CD",
                    "DVD"
                ]
            },
            "Floppy": {
                "Name": "Virtual Removable Media",
                "MediaTypes": [
                    "Floppy",
                    "USBStick"
                ]
            }
        }
    }

    def setUp(self):
        super().setUp()
        with mock.patch('sushy_tools.emulator.memoize.PersistentDict',
                        return_value={}, autospec=True):
            self.test_driver = vmedia.StaticDriver(self.CONFIG,
                                                   mock.MagicMock())

    def test_devices(self):
        devices = self.test_driver.devices
        self.assertEqual(['Cd', 'Floppy'], sorted(devices))

    def test_get_device_name(self):
        device_name = self.test_driver.get_device_name(
            self.UUID, 'Cd')
        self.assertEqual('Virtual CD', device_name)

    def test_get_device_media_types(self):
        media_types = self.test_driver.get_device_media_types(
            self.UUID, 'Cd')
        self.assertEqual(['CD', 'DVD'], media_types)

    def test_get_device_image_info(self):
        dev_info = self.test_driver.get_device_image_info(
            self.UUID, 'Cd')
        expected = ('', '', False, False, '', '', False)
        self.assertEqual(expected, dev_info)

    def test_update_device_info(self):
        dev_info = self.test_driver.get_device_image_info(self.UUID, 'Cd')
        self.assertFalse(dev_info.verify)

        self.test_driver.update_device_info(self.UUID, 'Cd', verify=True)
        dev_info = self.test_driver.get_device_image_info(self.UUID, 'Cd')
        self.assertTrue(dev_info.verify)

    @mock.patch.object(vmedia.StaticDriver, '_get_device', autospec=True)
    @mock.patch.object(builtins, 'open', autospec=True)
    @mock.patch.object(vmedia.os, 'rename', autospec=True)
    @mock.patch.object(vmedia, 'tempfile', autospec=True)
    @mock.patch.object(vmedia, 'requests', autospec=True)
    def test_insert_image(self, mock_requests, mock_tempfile, mock_rename,
                          mock_open, mock_get_device):
        device_info = {}
        mock_get_device.return_value = device_info

        mock_tempfile.mkdtemp.return_value = '/alphabet/soup'
        mock_tempfile.gettempdir.return_value = '/tmp'
        mock_tmp_file = (mock_tempfile.NamedTemporaryFile
                         .return_value.__enter__.return_value)
        mock_tmp_file.name = 'alphabet.soup'
        mock_rsp = mock_requests.get.return_value.__enter__.return_value
        mock_rsp.headers = {
            'content-disposition': 'attachment; filename="fish.iso"'
        }
        mock_rsp.status_code = 200

        local_file = self.test_driver.insert_image(
            self.UUID, 'Cd', 'http://fish.it/red.iso', inserted=True,
            write_protected=False)

        self.assertEqual('/alphabet/soup/fish.iso', local_file)
        mock_requests.get.assert_called_once_with(
            'http://fish.it/red.iso', stream=True, verify=False, auth=None)
        mock_open.assert_called_once_with(mock.ANY, 'wb')
        mock_rename.assert_called_once_with(
            'alphabet.soup', '/alphabet/soup/fish.iso')

        self.assertEqual('http://fish.it/red.iso', device_info['Image'])
        self.assertEqual('fish.iso', device_info['ImageName'])
        self.assertTrue(device_info['Inserted'])
        self.assertFalse(device_info['WriteProtected'])
        self.assertEqual('', device_info['UserName'])
        self.assertEqual('', device_info['Password'])
        self.assertEqual(local_file, device_info['_local_file'])

    @mock.patch.object(vmedia.StaticDriver, '_get_device', autospec=True)
    @mock.patch.object(builtins, 'open', autospec=True)
    @mock.patch.object(vmedia.os, 'rename', autospec=True)
    @mock.patch.object(vmedia, 'tempfile', autospec=True)
    @mock.patch.object(vmedia, 'requests', autospec=True)
    def test_insert_image_auth(self, mock_requests, mock_tempfile, mock_rename,
                               mock_open, mock_get_device):
        device_info = {}
        mock_get_device.return_value = device_info

        mock_tempfile.mkdtemp.return_value = '/alphabet/soup'
        mock_tempfile.gettempdir.return_value = '/tmp'
        mock_tmp_file = (mock_tempfile.NamedTemporaryFile
                         .return_value.__enter__.return_value)
        mock_tmp_file.name = 'alphabet.soup'
        mock_rsp = mock_requests.get.return_value.__enter__.return_value
        mock_rsp.headers = {
            'content-disposition': 'attachment; filename="fish.iso"'
        }
        mock_rsp.status_code = 200

        local_file = self.test_driver.insert_image(
            self.UUID, 'Cd', 'http://fish.it/red.iso', inserted=True,
            write_protected=False, username='Admin', password='Secret')

        self.assertEqual('/alphabet/soup/fish.iso', local_file)
        mock_requests.get.assert_called_once_with(
            'http://fish.it/red.iso', stream=True, verify=False,
            auth=('Admin', 'Secret'))
        mock_open.assert_called_once_with(mock.ANY, 'wb')
        mock_rename.assert_called_once_with(
            'alphabet.soup', '/alphabet/soup/fish.iso')

        self.assertEqual('http://fish.it/red.iso', device_info['Image'])
        self.assertEqual('fish.iso', device_info['ImageName'])
        self.assertTrue(device_info['Inserted'])
        self.assertFalse(device_info['WriteProtected'])
        self.assertEqual('Admin', device_info['UserName'])
        self.assertEqual('Secret', device_info['Password'])
        self.assertEqual(local_file, device_info['_local_file'])

    @mock.patch.object(vmedia.StaticDriver, '_get_device', autospec=True)
    @mock.patch.object(builtins, 'open', autospec=True)
    @mock.patch.object(vmedia.os, 'rename', autospec=True)
    @mock.patch.object(vmedia, 'tempfile', autospec=True)
    @mock.patch.object(vmedia, 'requests', autospec=True)
    def test_insert_image_no_local_file(self, mock_requests, mock_tempfile,
                                        mock_rename, mock_open,
                                        mock_get_device):
        device_info = {}
        mock_get_device.return_value = device_info

        mock_tempfile.mkdtemp.return_value = '/alphabet/soup'
        mock_tempfile.gettempdir.return_value = '/tmp'
        mock_tmp_file = (mock_tempfile.NamedTemporaryFile
                         .return_value.__enter__.return_value)
        mock_tmp_file.name = 'alphabet.soup'
        mock_rsp = mock_requests.get.return_value.__enter__.return_value
        mock_rsp.headers = {}
        mock_rsp.status_code = 200

        local_file = self.test_driver.insert_image(
            self.UUID, 'Cd', 'http://fish.it/red.iso', inserted=True,
            write_protected=False)

        self.assertEqual('/alphabet/soup/red.iso', local_file)
        mock_requests.get.assert_called_once_with(
            'http://fish.it/red.iso', stream=True, verify=False, auth=None)
        mock_open.assert_called_once_with(mock.ANY, 'wb')
        mock_rename.assert_called_once_with(
            'alphabet.soup', '/alphabet/soup/red.iso')

        self.assertEqual('http://fish.it/red.iso', device_info['Image'])
        self.assertEqual('red.iso', device_info['ImageName'])
        self.assertTrue(device_info['Inserted'])
        self.assertFalse(device_info['WriteProtected'])

    @mock.patch.object(vmedia.StaticDriver, '_get_device', autospec=True)
    @mock.patch.object(builtins, 'open', autospec=True)
    @mock.patch.object(vmedia.os, 'rename', autospec=True)
    @mock.patch.object(vmedia, 'tempfile', autospec=True)
    @mock.patch.object(vmedia, 'requests', autospec=True)
    def test_insert_image_full_url_v6(self, mock_requests, mock_tempfile,
                                      mock_rename, mock_open,
                                      mock_get_device):
        device_info = {}
        mock_get_device.return_value = device_info

        mock_tempfile.mkdtemp.return_value = '/alphabet/soup'
        mock_tempfile.gettempdir.return_value = '/tmp'
        mock_tmp_file = (mock_tempfile.NamedTemporaryFile
                         .return_value.__enter__.return_value)
        mock_tmp_file.name = 'alphabet.soup'
        mock_rsp = mock_requests.get.return_value.__enter__.return_value
        mock_rsp.headers = {}
        mock_rsp.status_code = 200

        full_url = 'http://[::2]:80/redfish/boot-abc?filename=tmp.iso'
        local_file = self.test_driver.insert_image(
            self.UUID, 'Cd', full_url,
            inserted=True, write_protected=False)

        self.assertEqual('/alphabet/soup/boot-abc', local_file)
        mock_requests.get.assert_called_once_with(full_url, stream=True,
                                                  verify=False, auth=None)
        mock_open.assert_called_once_with(mock.ANY, 'wb')
        mock_rename.assert_called_once_with(
            'alphabet.soup', '/alphabet/soup/boot-abc')

        self.assertEqual(full_url, device_info['Image'])
        self.assertEqual('boot-abc', device_info['ImageName'])
        self.assertTrue(device_info['Inserted'])
        self.assertFalse(device_info['WriteProtected'])

    @mock.patch.object(vmedia.StaticDriver, '_get_device', autospec=True)
    @mock.patch.object(builtins, 'open', autospec=True)
    @mock.patch.object(vmedia.os, 'rename', autospec=True)
    @mock.patch.object(vmedia, 'tempfile', autospec=True)
    @mock.patch.object(vmedia, 'requests', autospec=True)
    def test_insert_image_verify_ssl(self, mock_requests, mock_tempfile,
                                     mock_rename, mock_open,
                                     mock_get_device):
        device_info = {}
        mock_get_device.return_value = device_info

        mock_tempfile.mkdtemp.return_value = '/alphabet/soup'
        mock_tempfile.gettempdir.return_value = '/tmp'
        mock_tmp_file = (mock_tempfile.NamedTemporaryFile
                         .return_value.__enter__.return_value)
        mock_tmp_file.name = 'alphabet.soup'
        mock_rsp = mock_requests.get.return_value.__enter__.return_value
        mock_rsp.headers = {
            'content-disposition': 'attachment; filename="fish.iso"'
        }
        mock_rsp.status_code = 200

        ssl_conf_key = 'SUSHY_EMULATOR_VMEDIA_VERIFY_SSL'
        default_ssl_verify = self.test_driver._config.get(ssl_conf_key)
        try:
            self.test_driver._config[ssl_conf_key] = True
            local_file = self.test_driver.insert_image(
                self.UUID, 'Cd', 'https://fish.it/red.iso', inserted=True,
                write_protected=False)
        finally:
            self.test_driver._config[ssl_conf_key] = default_ssl_verify

        self.assertEqual('/alphabet/soup/fish.iso', local_file)
        mock_requests.get.assert_called_once_with(
            'https://fish.it/red.iso', stream=True, verify=True, auth=None)
        mock_open.assert_called_once_with(mock.ANY, 'wb')
        mock_rename.assert_called_once_with(
            'alphabet.soup', '/alphabet/soup/fish.iso')

        self.assertEqual('https://fish.it/red.iso', device_info['Image'])
        self.assertEqual('fish.iso', device_info['ImageName'])
        self.assertTrue(device_info['Inserted'])
        self.assertFalse(device_info['WriteProtected'])
        self.assertEqual(local_file, device_info['_local_file'])

    @mock.patch.object(vmedia.StaticDriver, '_get_device', autospec=True)
    @mock.patch.object(builtins, 'open', autospec=True)
    @mock.patch.object(vmedia.os, 'rename', autospec=True)
    @mock.patch.object(vmedia, 'tempfile', autospec=True)
    @mock.patch.object(vmedia, 'requests', autospec=True)
    def test_insert_image_verify_ssl_changed(self, mock_requests,
                                             mock_tempfile,
                                             mock_rename, mock_open,
                                             mock_get_device):
        device_info = {'Verify': True}
        mock_get_device.return_value = device_info

        mock_tempfile.mkdtemp.return_value = '/alphabet/soup'
        mock_tempfile.gettempdir.return_value = '/tmp'
        mock_tmp_file = (mock_tempfile.NamedTemporaryFile
                         .return_value.__enter__.return_value)
        mock_tmp_file.name = 'alphabet.soup'
        mock_rsp = mock_requests.get.return_value.__enter__.return_value
        mock_rsp.headers = {
            'content-disposition': 'attachment; filename="fish.iso"'
        }
        mock_rsp.status_code = 200

        local_file = self.test_driver.insert_image(
            self.UUID, 'Cd', 'https://fish.it/red.iso', inserted=True,
            write_protected=False)

        self.assertEqual('/alphabet/soup/fish.iso', local_file)
        mock_requests.get.assert_called_once_with(
            'https://fish.it/red.iso', stream=True, verify=True, auth=None)
        mock_open.assert_called_once_with(mock.ANY, 'wb')
        mock_rename.assert_called_once_with(
            'alphabet.soup', '/alphabet/soup/fish.iso')

        self.assertEqual('https://fish.it/red.iso', device_info['Image'])
        self.assertEqual('fish.iso', device_info['ImageName'])
        self.assertTrue(device_info['Inserted'])
        self.assertFalse(device_info['WriteProtected'])
        self.assertEqual(local_file, device_info['_local_file'])

    @mock.patch.object(vmedia.StaticDriver, '_get_device', autospec=True)
    @mock.patch.object(builtins, 'open', autospec=True)
    @mock.patch.object(vmedia.os, 'rename', autospec=True)
    @mock.patch.object(vmedia, 'tempfile', autospec=True)
    @mock.patch.object(vmedia, 'requests', autospec=True)
    def test_insert_image_verify_ssl_custom(self, mock_requests,
                                            mock_tempfile,
                                            mock_rename, mock_open,
                                            mock_get_device):
        device_info = {'Verify': True,
                       'Certificate': {'String': 'abcd'}}
        mock_get_device.return_value = device_info

        mock_tempfile.mkdtemp.return_value = '/alphabet/soup'
        mock_tempfile.gettempdir.return_value = '/tmp'
        mock_tmp_file = (mock_tempfile.NamedTemporaryFile
                         .return_value.__enter__.return_value)
        mock_tmp_file.name = 'alphabet.soup'
        mock_rsp = mock_requests.get.return_value.__enter__.return_value
        mock_rsp.headers = {
            'content-disposition': 'attachment; filename="fish.iso"'
        }
        mock_rsp.status_code = 200

        local_file = self.test_driver.insert_image(
            self.UUID, 'Cd', 'https://fish.it/red.iso', inserted=True,
            write_protected=False)

        self.assertEqual('/alphabet/soup/fish.iso', local_file)
        mock_requests.get.assert_called_once_with(
            'https://fish.it/red.iso', stream=True,
            verify=mock_tempfile.NamedTemporaryFile.return_value.name,
            auth=None)
        mock_open.assert_called_once_with(mock.ANY, 'wb')
        mock_rename.assert_called_once_with(
            'alphabet.soup', '/alphabet/soup/fish.iso')

        self.assertEqual('https://fish.it/red.iso', device_info['Image'])
        self.assertEqual('fish.iso', device_info['ImageName'])
        self.assertTrue(device_info['Inserted'])
        self.assertFalse(device_info['WriteProtected'])
        self.assertEqual(local_file, device_info['_local_file'])

    @mock.patch.object(vmedia.StaticDriver, '_get_device', autospec=True)
    @mock.patch.object(builtins, 'open', autospec=True)
    @mock.patch.object(vmedia.os, 'rename', autospec=True)
    @mock.patch.object(vmedia, 'tempfile', autospec=True)
    @mock.patch.object(vmedia, 'requests', autospec=True)
    def test_insert_image_fail(self, mock_requests, mock_tempfile, mock_rename,
                               mock_open, mock_get_device):
        device_info = {}
        mock_get_device.return_value = device_info

        mock_tempfile.mkdtemp.return_value = '/alphabet/soup'
        mock_tempfile.gettempdir.return_value = '/tmp'
        mock_tmp_file = (mock_tempfile.NamedTemporaryFile
                         .return_value.__enter__.return_value)
        mock_tmp_file.name = 'alphabet.soup'
        mock_rsp = mock_requests.get.return_value.__enter__.return_value
        mock_rsp.headers = {
            'content-disposition': 'attachment; filename="fish.iso"'
        }
        mock_rsp.status_code = 401

        self.assertRaises(error.FishyError,
                          self.test_driver.insert_image,
                          self.UUID, 'Cd', 'http://fish.it/red.iso',
                          inserted=True, write_protected=False)
        mock_requests.get.assert_called_once_with(
            'http://fish.it/red.iso', stream=True, auth=None, verify=False)
        mock_open.assert_not_called()
        self.assertEqual({}, device_info)

    @mock.patch.object(vmedia.StaticDriver, '_get_device', autospec=True)
    @mock.patch.object(vmedia.os, 'unlink', autospec=True)
    def test_eject_image(self, mock_unlink, mock_get_device):
        device_info = {
            '_local_file': '/tmp/fish.iso'
        }
        mock_get_device.return_value = device_info

        self.test_driver.eject_image(self.UUID, 'Cd')

        self.assertEqual('', device_info['Image'])
        self.assertEqual('', device_info['ImageName'])
        self.assertFalse(device_info['Inserted'])
        self.assertFalse(device_info['WriteProtected'])

        mock_unlink.assert_called_once_with('/tmp/fish.iso')

    @mock.patch.object(vmedia.StaticDriver, '_get_device', autospec=True)
    def test_list_certificates(self, mock_get_device):
        mock_get_device.return_value = {
            'Certificate': {'Type': 'PEM', 'String': 'abcd'},
        }
        result = self.test_driver.list_certificates(self.UUID, 'Cd')
        self.assertEqual([vmedia.Certificate('Default', 'abcd', 'PEM')],
                         result)

    def test_list_certificates_empty(self):
        result = self.test_driver.list_certificates(self.UUID, 'Cd')
        self.assertEqual([], result)

    @mock.patch.object(vmedia.StaticDriver, '_get_device', autospec=True)
    def test_add_certificate(self, mock_get_device):
        device_info = {}
        mock_get_device.return_value = device_info

        result = self.test_driver.add_certificate(self.UUID, 'Cd',
                                                  'abcd', 'PEM')
        self.assertEqual(vmedia.Certificate('Default', 'abcd', 'PEM'), result)
        self.assertEqual({'Certificate': {'Type': 'PEM', 'String': 'abcd'}},
                         device_info)

    @mock.patch.object(vmedia.StaticDriver, '_get_device', autospec=True)
    def test_add_certificate_exists(self, mock_get_device):
        device_info = {
            'Certificate': {'Type': 'PEM', 'String': 'abcd'},
        }
        mock_get_device.return_value = device_info

        self.assertRaises(error.FishyError,
                          self.test_driver.add_certificate,
                          self.UUID, 'Cd', 'defg', 'PEM')

    @mock.patch.object(vmedia.StaticDriver, '_get_device', autospec=True)
    def test_replace_certificate(self, mock_get_device):
        device_info = {
            'Certificate': {'Type': 'PEM', 'String': 'abcd'},
        }
        mock_get_device.return_value = device_info

        result = self.test_driver.replace_certificate(self.UUID, 'Cd',
                                                      'Default', 'defg', 'PEM')
        self.assertEqual(vmedia.Certificate('Default', 'defg', 'PEM'), result)
        self.assertEqual({'Certificate': {'Type': 'PEM', 'String': 'defg'}},
                         device_info)

    @mock.patch.object(vmedia.StaticDriver, '_get_device', autospec=True)
    def test_replace_certificate_wrong_id(self, mock_get_device):
        device_info = {
            'Certificate': {'Type': 'PEM', 'String': 'abcd'},
        }
        mock_get_device.return_value = device_info

        self.assertRaises(error.NotFound,
                          self.test_driver.replace_certificate,
                          self.UUID, 'Cd', 'Other', 'defg', 'PEM')

    @mock.patch.object(vmedia.StaticDriver, '_get_device', autospec=True)
    def test_replace_certificate_not_found(self, mock_get_device):
        device_info = {}
        mock_get_device.return_value = device_info

        self.assertRaises(error.NotFound,
                          self.test_driver.replace_certificate,
                          self.UUID, 'Cd', 'Default', 'defg', 'PEM')

    @mock.patch.object(vmedia.StaticDriver, '_get_device', autospec=True)
    def test_delete_certificate(self, mock_get_device):
        device_info = {
            'Certificate': {'Type': 'PEM', 'String': 'abcd'},
        }
        mock_get_device.return_value = device_info

        self.test_driver.delete_certificate(self.UUID, 'Cd', 'Default')
        self.assertEqual({}, device_info)

    @mock.patch.object(vmedia.StaticDriver, '_get_device', autospec=True)
    def test_delete_certificate_not_found(self, mock_get_device):
        device_info = {}
        mock_get_device.return_value = device_info

        self.assertRaises(error.NotFound,
                          self.test_driver.delete_certificate,
                          self.UUID, 'Cd', 'Default')



#    Licensed under the Apache License, Version 2.0 (the "License"); you may
#    not use this file except in compliance with the License. You may obtain
#    a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#    WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#    License for the specific language governing permissions and limitations
#    under the License.

import tempfile
from unittest import mock

from oslotest import base

from sushy_tools.emulator import main
from sushy_tools import error


def patch_resource(name):
    def decorator(func):
        return mock.patch.object(main.Application, name,
                                 new_callable=mock.PropertyMock)(func)
    return decorator


class EmulatorTestCase(base.BaseTestCase):

    name = 'QEmu-fedora-i686'
    uuid = 'c7a5fdbd-cdaf-9455-926a-d65c16db1809'

    def setUp(self):
        self.app = main.app.test_client()

        super(EmulatorTestCase, self).setUp()


class CommonTestCase(EmulatorTestCase):

    @patch_resource('systems')
    def test_error(self, systems_mock):
        systems_mock.return_value.get_power_state.side_effect = Exception(
            'Fish is dead')
        response = self.app.get('/redfish/v1/Systems/' + self.uuid)

        self.assertEqual(500, response.status_code)

    def test_root_resource(self):
        response = self.app.get('/redfish/v1/')
        self.assertEqual(200, response.status_code)
        self.assertEqual('RedvirtService', response.json['Id'])


TEST_PASSWD = \
    b"admin:$2y$05$mYl8KMwM94l4LR/sw1teIeA6P2u8gfX16e8wvT7NmGgAM5r9jgLl."


class AuthenticatedTestCase(base.BaseTestCase):

    def setUp(self):
        super().setUp()
        self.auth_file = tempfile.NamedTemporaryFile()
        self.auth_file.write(TEST_PASSWD)
        self.auth_file.flush()
        self.addCleanup(self.auth_file.close)
        app = main.Application()
        app.configure(
            extra_config={'SUSHY_EMULATOR_AUTH_FILE': self.auth_file.name})
        self.app = app.test_client()

    def test_root_resource(self):
        response = self.app.get('/redfish/v1/')
        # 404 because this application does not have any routes
        self.assertEqual(404, response.status_code, response.data)

    def test_authenticated_resource(self):
        response = self.app.get('/redfish/v1/Systems/',
                                auth=('admin', 'password'))
        self.assertEqual(404, response.status_code, response.data)

    def test_authentication_failed(self):
        response = self.app.get('/redfish/v1/Systems/')
        self.assertEqual(401, response.status_code, response.data)


class ChassisTestCase(EmulatorTestCase):

    @patch_resource('chassis')
    def test_chassis_collection_resource(self, chassis_mock):
        chassis_mock.return_value.chassis = ['chassis0', 'chassis1']
        response = self.app.get('/redfish/v1/Chassis')
        self.assertEqual(200, response.status_code)
        self.assertEqual({'@odata.id': '/redfish/v1/Chassis/chassis0'},
                         response.json['Members'][0])
        self.assertEqual({'@odata.id': '/redfish/v1/Chassis/chassis1'},
                         response.json['Members'][1])

    @patch_resource('indicators')
    @patch_resource('systems')
    @patch_resource('managers')
    @patch_resource('chassis')
    def test_chassis_resource_get(self, chassis_mock, managers_mock,
                                  systems_mock, indicators_mock):
        chassis_mock = chassis_mock.return_value
        chassis_mock.chassis = ['xxxx-yyyy-zzzz']
        chassis_mock.uuid.return_value = 'xxxx-yyyy-zzzz'
        chassis_mock.name.return_value = 'name'
        managers_mock.return_value.managers = ['man1']
        systems_mock.return_value.systems = ['sys1']
        indicators_mock.return_value.get_indicator_state.return_value = 'Off'

        response = self.app.get('/redfish/v1/Chassis/xxxx-yyyy-zzzz')

        self.assertEqual(200, response.status_code)
        self.assertEqual('xxxx-yyyy-zzzz', response.json['Id'])
        self.assertEqual('xxxx-yyyy-zzzz', response.json['UUID'])
        self.assertEqual('Off', response.json['IndicatorLED'])
        self.assertEqual(
            {'@odata.id': '/redfish/v1/Chassis/xxxx-yyyy-zzzz/Thermal'},
            response.json['Thermal'])
        self.assertEqual([{'@odata.id': '/redfish/v1/Systems/sys1'}],
                         response.json['Links']['ComputerSystems'])
        self.assertEqual([{'@odata.id': '/redfish/v1/Managers/man1'}],
                         response.json['Links']['ManagedBy'])
        self.assertEqual([{'@odata.id': '/redfish/v1/Managers/man1'}],
                         response.json['Links']['ManagersInChassis'])

    @patch_resource('systems')
    @patch_resource('chassis')
    def test_chassis_thermal(self, chassis_mock, systems_mock):
        chassis_mock = chassis_mock.return_value
        chassis_mock.chassis = [self.uuid]
        chassis_mock.uuid.return_value = self.uuid
        systems_mock.return_value.systems = ['sys1']

        response = self.app.get('/redfish/v1/Chassis/xxxx-yyyy-zzzz/Thermal')

        self.assertEqual(200, response.status_code)
        self.assertEqual('Thermal', response.json['Id'])
        self.assertEqual(
            '/redfish/v1/Chassis/xxxx-yyyy-zzzz/Thermal#/Temperatures/0',
            response.json['Temperatures'][0]['@odata.id'])
        self.assertEqual(
            {'@odata.id': '/redfish/v1/Systems/sys1/Processors/CPU'},
            response.json['Temperatures'][0]['RelatedItem'][0])
        self.assertEqual(
            '/redfish/v1/Chassis/xxxx-yyyy-zzzz/Thermal#/Fans/0',
            response.json['Fans'][0]['@odata.id'])
        self.assertEqual(
            {'@odata.id': '/redfish/v1/Chassis/xxxx-yyyy-zzzz'},
            response.json['Fans'][0]['RelatedItem'][0])

    @patch_resource('indicators')
    @patch_resource('chassis')
    def test_chassis_indicator_set_ok(self, chassis_mock, indicators_mock):
        chassis_mock.return_value.uuid.return_value = self.uuid

        data = {'IndicatorLED': 'Off'}
        response = self.app.patch('/redfish/v1/Chassis/xxxx-yyyy-zzzz',
                                  json=data)
        self.assertEqual(204, response.status_code)
        set_indicator_state = indicators_mock.return_value.set_indicator_state
        set_indicator_state.assert_called_once_with(self.uuid, 'Off')

    @patch_resource('indicators')
    @patch_resource('chassis')
    def test_chassis_indicator_set_fail(self, chassis_mock, indicators_mock):
        set_indicator_state = indicators_mock.return_value.set_indicator_state
        set_indicator_state.side_effect = error.FishyError
        data = {'IndicatorLED': 'Blah'}
        response = self.app.patch('/redfish/v1/Chassis/xxxx-yyyy-zzzz',
                                  json=data)
        self.assertEqual(500, response.status_code)


class ManagersTestCase(EmulatorTestCase):

    @patch_resource('managers')
    def test_manager_collection_resource(self, managers_mock):
        type(managers_mock.return_value).managers = mock.PropertyMock(
            return_value=['bmc0', 'bmc1'])
        response = self.app.get('/redfish/v1/Managers')
        self.assertEqual(200, response.status_code)
        self.assertEqual({'@odata.id': '/redfish/v1/Managers/bmc0'},
                         response.json['Members'][0])
        self.assertEqual({'@odata.id': '/redfish/v1/Managers/bmc1'},
                         response.json['Members'][1])

    @patch_resource('managers')
    def test_manager_resource_get(self, managers_mock):
        managers_mock = managers_mock.return_value
        managers_mock.managers = ['xxxx-yyyy-zzzz']
        managers_mock.get_manager.return_value = {
            'UUID': 'xxxx-yyyy-zzzz',
            'Name': 'name',
            'Id': 'xxxx-yyyy-zzzz',
        }
        managers_mock.get_managed_systems.return_value = ['xxx']
        managers_mock.get_managed_chassis.return_value = ['chassis0']

        response = self.app.get('/redfish/v1/Managers/xxxx-yyyy-zzzz')

        self.assertEqual(200, response.status_code, response.json)
        self.assertEqual('xxxx-yyyy-zzzz', response.json['Id'])
        self.assertEqual('xxxx-yyyy-zzzz', response.json['UUID'])
        self.assertIsNone(response.json['ServiceEntryPointUUID'])
        self.assertEqual([{'@odata.id': '/redfish/v1/Systems/xxx'}],
                         response.json['Links']['ManagerForServers'])
        self.assertEqual([{'@odata.id': '/redfish/v1/Chassis/chassis0'}],
                         response.json['Links']['ManagerForChassis'])


class SystemsTestCase(EmulatorTestCase):

    @patch_resource('systems')
    def test_system_collection_resource(self, systems_mock):
        type(systems_mock.return_value).systems = mock.PropertyMock(
            return_value=['host0', 'host1'])
        response = self.app.get('/redfish/v1/Systems')
        self.assertEqual(200, response.status_code)
        self.assertEqual({'@odata.id': '/redfish/v1/Systems/host0'},
                         response.json['Members'][0])
        self.assertEqual({'@odata.id': '/redfish/v1/Systems/host1'},
                         response.json['Members'][1])

    @patch_resource('indicators')
    @patch_resource('chassis')
    @patch_resource('managers')
    @patch_resource('systems')
    def test_system_resource_get(self, systems_mock, managers_mock,
                                 chassis_mock, indicators_mock):
        systems_mock = systems_mock.return_value
        systems_mock.uuid.return_value = 'zzzz-yyyy-xxxx'
        systems_mock.get_power_state.return_value = 'On'
        systems_mock.get_total_memory.return_value = 1
        systems_mock.get_total_cpus.return_value = 2
        systems_mock.get_boot_device.return_value = 'Cd'
        systems_mock.get_boot_mode.return_value = 'Legacy'
        managers_mock.return_value.get_managers_for_system.return_value = [
            'aaaa-bbbb-cccc']
        chassis_mock.return_value.chassis = ['chassis0']
        indicators_mock.return_value.get_indicator_state.return_value = 'Off'

        response = self.app.get('/redfish/v1/Systems/xxxx-yyyy-zzzz')

        self.assertEqual(200, response.status_code)
        self.assertEqual('xxxx-yyyy-zzzz', response.json['Id'])
        self.assertEqual('zzzz-yyyy-xxxx', response.json['UUID'])
        self.assertEqual('On', response.json['PowerState'])
        self.assertEqual('Off', response.json['IndicatorLED'])
        self.assertEqual(
            1, response.json['MemorySummary']['TotalSystemMemoryGiB'])
        self.assertEqual(2, response.json['ProcessorSummary']['Count'])
        self.assertEqual(
            'Cd', response.json['Boot']['BootSourceOverrideTarget'])
        self.assertEqual(
            'Legacy', response.json['Boot']['BootSourceOverrideMode'])
        self.assertEqual(
            [{'@odata.id': '/redfish/v1/Managers/aaaa-bbbb-cccc'}],
            response.json['Links']['ManagedBy'])
        self.assertEqual(
            [{'@odata.id': '/redfish/v1/Chassis/chassis0'}],
            response.json['Links']['Chassis'])

    @patch_resource('systems')
    def test_system_resource_patch(self, systems_mock):
        data = {'Boot': {'BootSourceOverrideTarget': 'Cd'}}
        response = self.app.patch('/redfish/v1/Systems/xxxx-yyyy-zzzz',
                                  json=data)
        self.assertEqual(204, response.status_code)
        set_boot_device = systems_mock.return_value.set_boot_device
        set_boot_device.assert_called_once_with('xxxx-yyyy-zzzz', 'Cd')

    @patch_resource('systems')
    def test_system_reset_action(self, systems_mock):
        set_power_state = systems_mock.return_value.set_power_state
        for reset_type in ('On', 'ForceOn', 'ForceOff', 'GracefulShutdown',
                           'GracefulRestart', 'ForceRestart', 'Nmi'):
            set_power_state.reset_mock()
            data = {'ResetType': reset_type}
            response = self.app.post(
                '/redfish/v1/Systems/xxxx-yyyy-zzzz/Actions/'
                'ComputerSystem.Reset',
                json=data)
            self.assertEqual(204, response.status_code)
            set_power_state.assert_called_once_with('xxxx-yyyy-zzzz',
                                                    reset_type)

    @patch_resource('indicators')
    @patch_resource('systems')
    def test_system_indicator_set_ok(self, systems_mock, indicators_mock):
        systems_mock.return_value.uuid.return_value = self.uuid

        data = {'IndicatorLED': 'Off'}
        response = self.app.patch('/redfish/v1/Systems/xxxx-yyyy-zzzz',
                                  json=data)
        self.assertEqual(204, response.status_code)
        set_indicator_state = indicators_mock.return_value.set_indicator_state
        set_indicator_state.assert_called_once_with(self.uuid, 'Off')

    @patch_resource('indicators')
    @patch_resource('systems')
    def test_system_indicator_set_fail(self, systems_mock, indicators_mock):
        set_indicator_state = indicators_mock.return_value.set_indicator_state
        set_indicator_state.side_effect = error.FishyError
        data = {'IndicatorLED': 'Blah'}
        response = self.app.patch('/redfish/v1/Systems/xxxx-yyyy-zzzz',
                                  json=data)
        self.assertEqual(500, response.status_code)


@patch_resource('systems')
class BiosTestCase(EmulatorTestCase):

    def test_get_bios(self, systems_mock):
        systems_mock.return_value.get_bios.return_value = {
            "attribute 1": "value 1",
            "attribute 2": "value 2"
        }
        response = self.app.get('/redfish/v1/Systems/' + self.uuid + '/BIOS')

        self.assertEqual(200, response.status_code)
        self.assertEqual('BIOS', response.json['Id'])
        self.assertEqual({"attribute 1": "value 1",
                          "attribute 2": "value 2"},
                         response.json['Attributes'])

    def test_get_bios_existing(self, systems_mock):
        systems_mock.return_value.get_bios.return_value = {
            "attribute 1": "value 1",
            "attribute 2": "value 2"
        }
        response = self.app.get(
            '/redfish/v1/Systems/' + self.uuid + '/BIOS/Settings')

        self.assertEqual(200, response.status_code)
        self.assertEqual('Settings', response.json['Id'])
        self.assertEqual(
            {"attribute 1": "value 1",
             "attribute 2": "value 2"},
            response.json['Attributes'])

    def test_bios_settings_patch(self, systems_mock):
        data = {'Attributes': {'key': 'value'}}
        response = self.app.patch(
            '/redfish/v1/Systems/xxxx-yyyy-zzzz/BIOS/Settings',
            json=data)
        self.assertEqual(204, response.status_code)
        systems_mock.return_value.set_bios.assert_called_once_with(
            'xxxx-yyyy-zzzz', {'key': 'value'})

    def test_set_bios(self, systems_mock):
        data = {'Attributes': {'key': 'value'}}
        response = self.app.patch(
            '/redfish/v1/Systems/xxxx-yyyy-zzzz/BIOS/Settings',
            json=data)

        self.assertEqual(204, response.status_code)
        systems_mock.return_value.set_bios.assert_called_once_with(
            'xxxx-yyyy-zzzz', data['Attributes'])

    def test_reset_bios(self, systems_mock):
        response = self.app.post('/redfish/v1/Systems/%s/BIOS/Actions/'
                                 'Bios.ResetBios' % self.uuid)
        self.assertEqual(204, response.status_code)
        systems_mock.return_value.reset_bios.assert_called_once_with(self.uuid)


@patch_resource('systems')
class EthernetInterfacesTestCase(EmulatorTestCase):

    def test_ethernet_interfaces_collection(self, systems_mock):
        systems_mock.return_value.get_nics.return_value = [
            {'id': 'nic1', 'mac': '52:54:00:4e:5d:37'},
            {'id': 'nic2', 'mac': '00:11:22:33:44:55'}]
        response = self.app.get('redfish/v1/Systems/%s/EthernetInterfaces'
                                % self.uuid)

        self.assertEqual(200, response.status_code)
        self.assertEqual('Ethernet Interface Collection',
                         response.json['Name'])
        self.assertEqual(2, response.json['Members@odata.count'])
        self.assertEqual(['/redfish/v1/Systems/%s/EthernetInterfaces/nic1'
                          % self.uuid,
                          '/redfish/v1/Systems/%s/EthernetInterfaces/nic2'
                          % self.uuid],
                         [m['@odata.id'] for m in response.json['Members']])

    def test_ethernet_interfaces_collection_empty(self, systems_mock):
        systems_mock.return_value.get_nics.return_value = []
        response = self.app.get('redfish/v1/Systems/%s/EthernetInterfaces'
                                % self.uuid)

        self.assertEqual(200, response.status_code)
        self.assertEqual('Ethernet Interface Collection',
                         response.json['Name'])
        self.assertEqual(0, response.json['Members@odata.count'])
        self.assertEqual([], response.json['Members'])

    def test_ethernet_interface(self, systems_mock):
        systems_mock.return_value.get_nics.return_value = [
            {'id': 'nic1', 'mac': '52:54:00:4e:5d:37'},
            {'id': 'nic2', 'mac': '00:11:22:33:44:55'}]
        response = self.app.get('/redfish/v1/Systems/%s/EthernetInterfaces/'
                                'nic2' % self.uuid)

        self.assertEqual(200, response.status_code)
        self.assertEqual('nic2', response.json['Id'])
        self.assertEqual('VNIC nic2', response.json['Name'])
        self.assertEqual('00:11:22:33:44:55',
                         response.json['PermanentMACAddress'])
        self.assertEqual('00:11:22:33:44:55',
                         response.json['MACAddress'])
        self.assertEqual('/redfish/v1/Systems/%s/EthernetInterfaces/nic2'
                         % self.uuid,
                         response.json['@odata.id'])

    def test_ethernet_interface_not_found(self, systems_mock):
        systems_mock.return_value.get_nics.return_value = [
            {'id': 'nic1', 'mac': '52:54:00:4e:5d:37'},
            {'id': 'nic2', 'mac': '00:11:22:33:44:55'}
        ]
        response = self.app.get('/redfish/v1/Systems/%s/EthernetInterfaces/'
                                'nic3' % self.uuid)

        self.assertEqual(404, response.status_code)


@patch_resource('systems')
class StorageTestCase(EmulatorTestCase):

    def test_simple_storage_collection(self, systems_mock):
        systems_mock = systems_mock.return_value
        systems_mock.get_simple_storage_collection.return_value = {
            'virtio': {
                'Id': 'virtio',
                'Name': 'virtio',
                'DeviceList': [
                    {
                        'Name': 'testVM1.img',
                        'CapacityBytes': 100000
                    },
                    {
                        'Name': 'sdb1',
                        'CapacityBytes': 150000
                    }
                ]
            },
            'ide': {
                'Id': 'ide',
                'Name': 'ide',
                'DeviceList': [
                    {
                        'Name': 'testVol1.img',
                        'CapacityBytes': 200000
                    },
                    {
                        'Name': 'blk-pool0-vol0',
                        'CapacityBytes': 300000
                    }
                ]
            }
        }
        response = self.app.get('redfish/v1/Systems/%s/SimpleStorage'
                                % self.uuid)

        self.assertEqual(200, response.status_code)
        self.assertEqual('Simple Storage Collection',
                         response.json['Name'])
        self.assertEqual(2, response.json['Members@odata.count'])
        self.assertEqual({'/redfish/v1/Systems/%s/SimpleStorage/virtio'
                          % self.uuid,
                          '/redfish/v1/Systems/%s/SimpleStorage/ide'
                          % self.uuid},
                         {m['@odata.id'] for m in response.json['Members']})

    def test_simple_storage_collection_empty(self, systems_mock):
        systems_mock = systems_mock.return_value
        systems_mock.get_simple_storage_collection.return_value = []
        response = self.app.get('redfish/v1/Systems/%s/SimpleStorage'
                                % self.uuid)

        self.assertEqual(200, response.status_code)
        self.assertEqual('Simple Storage Collection',
                         response.json['Name'])
        self.assertEqual(0, response.json['Members@odata.count'])
        self.assertEqual([], response.json['Members'])

    def test_simple_storage(self, systems_mock):
        systems_mock = systems_mock.return_value
        systems_mock.get_simple_storage_collection.return_value = {
            'virtio': {
                'Id': 'virtio',
                'Name': 'virtio',
                'DeviceList': [
                    {
                        'Name': 'testVM1.img',
                        'CapacityBytes': 100000
                    },
                    {
                        'Name': 'sdb1',
                        'CapacityBytes': 150000
                    }
                ]
            },
            'ide': {
                'Id': 'ide',
                'Name': 'ide',
                'DeviceList': [
                    {
                        'Name': 'testVol1.img',
                        'CapacityBytes': 200000
                    },
                    {
                        'Name': 'blk-pool0-vol0',
                        'CapacityBytes': 300000
                    }
                ]
            }
        }
        response = self.app.get('/redfish/v1/Systems/%s/SimpleStorage/virtio'
                                % self.uuid)

        self.assertEqual(200, response.status_code)
        self.assertEqual('virtio', response.json['Id'])
        self.assertEqual('virtio Controller', response.json['Name'])
        self.assertEqual('testVM1.img', response.json['Devices'][0]['Name'])
        self.assertEqual(100000, response.json['Devices'][0]['CapacityBytes'])
        self.assertEqual('sdb1', response.json['Devices'][1]['Name'])
        self.assertEqual(150000, response.json['Devices'][1]['CapacityBytes'])
        self.assertEqual('/redfish/v1/Systems/%s/SimpleStorage/virtio'
                         % self.uuid,
                         response.json['@odata.id'])

    def test_simple_storage_not_found(self, systems_mock):
        systems_mock = systems_mock.return_value
        systems_mock.get_simple_storage_collection.return_value = {
            'virtio': {
                'Id': 'virtio',
                'Name': 'virtio',
                'DeviceList': [
                    {
                        'Name': 'testVM1.img',
                        'CapacityBytes': 100000
                    },
                    {
                        'Name': 'sdb1',
                        'CapacityBytes': 150000
                    }
                ]
            },
            'ide': {
                'Id': 'ide',
                'Name': 'ide',
                'DeviceList': [
                    {
                        'Name': 'testVol1.img',
                        'CapacityBytes': 200000
                    },
                    {
                        'Name': 'blk-pool0-vol0',
                        'CapacityBytes': 300000
                    }
                ]
            }
        }
        response = self.app.get('/redfish/v1/Systems/%s/SimpleStorage/scsi'
                                % self.uuid)

        self.assertEqual(404, response.status_code)

    @patch_resource('storage')
    def test_storage_collection_resource(self, storage_mock, systems_mock):
        storage_mock.return_value.get_storage_col.return_value = [
            {
                "Id": "1",
                "Name": "Local Storage Controller",
                "StorageControllers": [
                    {
                        "MemberId": "0",
                        "Name": "Contoso Integrated RAID",
                        "SpeedGbps": 12
                    }
                ]
            }
        ]
        response = self.app.get('redfish/v1/Systems/vbmc-node/Storage')
        self.assertEqual(200, response.status_code)
        self.assertEqual({'@odata.id':
                          '/redfish/v1/Systems/vbmc-node/Storage/1'},
                         response.json['Members'][0])

    @patch_resource('storage')
    def test_storage_resource_get(self, storage_mock, systems_mock):
        storage_mock.return_value.get_storage_col.return_value = [
            {
                "Id": "1",
                "Name": "Local Storage Controller",
                "StorageControllers": [
                    {
                        "MemberId": "0",
                        "Name": "Contoso Integrated RAID",
                        "SpeedGbps": 12
                    }
                ]
            }
        ]
        response = self.app.get('/redfish/v1/Systems/vbmc-node/Storage/1')

        self.assertEqual(200, response.status_code)
        self.assertEqual('1', response.json['Id'])
        self.assertEqual('Local Storage Controller', response.json['Name'])
        stg_ctl = response.json['StorageControllers'][0]
        self.assertEqual("0", stg_ctl['MemberId'])
        self.assertEqual("Contoso Integrated RAID", stg_ctl['Name'])
        self.assertEqual(12, stg_ctl['SpeedGbps'])

    @patch_resource('drives')
    def test_drive_resource_get(self, drives_mock, systems_mock):
        drives_mock.return_value.get_drives.return_value = [
            {
                "Id": "32ADF365C6C1B7BD",
                "Name": "Drive Sample",
                "CapacityBytes": 899527000000,
                "Protocol": "SAS"
            }
        ]
        response = self.app.get('/redfish/v1/Systems/vbmc-node/Storage/1'
                                '/Drives/32ADF365C6C1B7BD')

        self.assertEqual(200, response.status_code)
        self.assertEqual('32ADF365C6C1B7BD', response.json['Id'])
        self.assertEqual('Drive Sample', response.json['Name'])
        self.assertEqual(899527000000, response.json['CapacityBytes'])
        self.assertEqual('SAS', response.json['Protocol'])

    @patch_resource('volumes')
    def test_volume_collection_get(self, volumes_mock, systems_mock):
        volumes_mock.return_value.get_volumes_col.return_value = [
            {
                "libvirtPoolName": "sushyPool",
                "libvirtVolName": "testVol",
                "Id": "1",
                "Name": "Sample Volume 1",
                "VolumeType": "Mirrored",
                "CapacityBytes": 23748
            }
        ]
        systems_mock = systems_mock.return_value
        systems_mock.find_or_create_storage_volume.return_value = "1"
        response = self.app.get('/redfish/v1/Systems/vmc-node/Storage/1/'
                                'Volumes')

        self.assertEqual(200, response.status_code)
        self.assertEqual({'@odata.id':
                          '/redfish/v1/Systems/vmc-node/Storage/1/Volumes/1'},
                         response.json['Members'][0])

    @patch_resource('volumes')
    def test_create_volume_post(self, volumes_mock, systems_mock):
        systems_mock = systems_mock.return_value
        systems_mock.find_or_create_storage_volume.return_value = "13087010612"
        data = {
            "Name": "Sample Volume 3",
            "VolumeType": "NonRedundant",
            "CapacityBytes": 23456
        }
        response = self.app.post('/redfish/v1/Systems/vmc-node/Storage/1/'
                                 'Volumes', json=data)

        self.assertEqual(201, response.status_code)
        self.assertIn('/redfish/v1/Systems/vmc-node/'
                      'Storage/1/Volumes/13087010612',
                      response.headers['Location'])

    @patch_resource('volumes')
    def test_volume_resource_get(self, volumes_mock, systems_mock):
        volumes_mock.return_value.get_volumes_col.return_value = [
            {
                "libvirtPoolName": "sushyPool",
                "libvirtVolName": "testVol",
                "Id": "1",
                "Name": "Sample Volume 1",
                "VolumeType": "Mirrored",
                "CapacityBytes": 23748
            }
        ]
        systems_mock = systems_mock.return_value
        systems_mock.find_or_create_storage_volume.return_value = "1"
        response = self.app.get('/redfish/v1/Systems/vbmc-node/Storage/1/'
                                'Volumes/1')

        self.assertEqual(200, response.status_code)
        self.assertEqual('1', response.json['Id'])
        self.assertEqual('Sample Volume 1', response.json['Name'])
        self.assertEqual('Mirrored', response.json['VolumeType'])
        self.assertEqual(23748, response.json['CapacityBytes'])


class RegistryTestCase(EmulatorTestCase):

    def test_registry_file_collection(self):
        response = self.app.get('/redfish/v1/Registries')
        self.assertEqual(200, response.status_code)
        self.assertEqual('/redfish/v1/Registries', response.json['@odata.id'])
        self.assertEqual(
            {'@odata.id': '/redfish/v1/Registries/Messages'},
            response.json['Members'][0])
        self.assertEqual(
            {'@odata.id':
                '/redfish/v1/Registries/BiosAttributeRegistry.v1_0_0'},
            response.json['Members'][1])

    def test_bios_attribute_registry_file(self):
        response = self.app.get(
            '/redfish/v1/Registries/BiosAttributeRegistry.v1_0_0')
        self.assertEqual(200, response.status_code)
        self.assertEqual('/redfish/v1/Registries/BiosAttributeRegistry.v1_0_0',
                         response.json['@odata.id'])
        self.assertEqual(
            {'Language': 'en',
             'Uri': '/redfish/v1/Systems/Bios/BiosRegistry'},
            response.json['Location'][0])

    def test_message_registry_file(self):
        response = self.app.get(
            '/redfish/v1/Registries/Messages')
        self.assertEqual(200, response.status_code)
        self.assertEqual('/redfish/v1/Registries/Messages',
                         response.json['@odata.id'])
        self.assertEqual({'Language': 'En',
                          'Uri': '/redfish/v1/Registries/Messages/Registry'},
                         response.json['Location'][0])

    def test_bios_registry(self):
        response = self.app.get(
            '/redfish/v1/Systems/Bios/BiosRegistry')
        self.assertEqual(200, response.status_code)
        self.assertEqual('/redfish/v1/Systems/Bios/BiosRegistry',
                         response.json['@odata.id'])
        self.assertEqual('BIOS Attribute Registry',
                         response.json['Name'])
        self.assertEqual('BiosAttributeRegistryP89.v1_0_0',
                         response.json['Id'])
        entries = response.json['RegistryEntries']
        self.assertEqual(len(entries['Attributes']), 10)
        self.assertEqual({'AttributeName': 'ProcTurboMode',
                          'CurrentValue': None,
                          'DisplayName': 'Turbo Boost',
                          'HelpText': 'Governs the Turbo Boost Technology. '
                          'This feature allows the processor cores to be '
                          'automatically clocked up in frequency beyond the '
                          'advertised processor speed.',
                          'Hidden': False,
                          'Immutable': False,
                          'ReadOnly': False,
                          'Type': 'Enumeration',
                          'Value': [{'ValueDisplayName': 'Enabled',
                                     'ValueName': 'Enabled'},
                                    {'ValueDisplayName': 'Disabled',
                                     'ValueName': 'Disabled'}],
                          'WarningText': None,
                          'WriteOnly': False},
                         entries['Attributes'][0])

    def test_message_registry(self):
        response = self.app.get('/redfish/v1/Registries/Messages/Registry')
        self.assertEqual(200, response.status_code)
        self.assertEqual('/redfish/v1/Registries/Messages/Registry',
                         response.json['@odata.id'])
        self.assertEqual('Message Registry', response.json['Name'])
        self.assertEqual('1.1.1', response.json['Id'])
        messages = response.json['Messages']
        self.assertEqual({"Description": "The command was successful.",
                          "Message": "The command was successful",
                          "Severity": "Informational",
                          "NumberOfArgs": 0,
                          "Resolution": "No response action is required."},
                         messages['BIOS001'])



# Copyright 2019 Red Hat, Inc.
# All Rights Reserved.
#
#    Licensed under the Apache License, Version 2.0 (the "License"); you may
#    not use this file except in compliance with the License. You may obtain
#    a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#    WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#    License for the specific language governing permissions and limitations
#    under the License.
from unittest import mock

from oslotest import base

from sushy_tools.emulator.resources.indicators import StaticDriver
from sushy_tools import error


class StaticDriverTestCase(base.BaseTestCase):

    UUID = "58893887-8974-2487-2389-841168418919"
    STATE = "Off"

    CONFIG = {
        'SUSHY_EMULATOR_INDICATOR_LEDS': {
            UUID: STATE
        }
    }

    def setUp(self):
        super().setUp()
        with mock.patch('sushy_tools.emulator.memoize.PersistentDict',
                        return_value={}, autospec=True):
            self.test_driver = StaticDriver(self.CONFIG, mock.MagicMock())

    def test_indicators(self):
        indicators = self.test_driver.indicators
        self.assertEqual([self.UUID], indicators)

    def test_get_indicator_state(self):
        state = self.test_driver.get_indicator_state(self.UUID)
        self.assertEqual('Off', state)

    def test_set_indicator_state_ok(self):
        self.test_driver.set_indicator_state(self.UUID, 'Lit')
        state = self.test_driver.get_indicator_state(self.UUID)
        self.assertEqual('Lit', state)

    def test_set_indicator_state_fail(self):
        self.assertRaises(
            error.FishyError,
            self.test_driver.set_indicator_state,
            self.UUID, 'Blah')



# Copyright 2019 Red Hat, Inc.
# All Rights Reserved.
#
#    Licensed under the Apache License, Version 2.0 (the "License"); you may
#    not use this file except in compliance with the License. You may obtain
#    a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#    WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#    License for the specific language governing permissions and limitations
#    under the License.

from sushy_tools.emulator import memoize
from sushy_tools.emulator.resources import base
from sushy_tools import error


class StaticDriver(base.DriverBase):
    """Redfish indicator LED simulator

    Maintain indicators states in memory. Does not light up
    anything.
    """

    def __init__(self, config, logger):
        super().__init__(config, logger)
        self._indicators = memoize.PersistentDict()
        if hasattr(self._indicators, 'make_permanent'):
            self._indicators.make_permanent(
                self._config.get('SUSHY_EMULATOR_STATE_DIR'), 'indicators')
        self._indicators.update(
            self._config.get('SUSHY_EMULATOR_INDICATOR_LEDS', {}))

    @property
    def driver(self):
        """Return human-friendly driver information

        :returns: driver information as `str`
        """
        return '<static-indicators>'

    @property
    def indicators(self):
        """Return available Redfish indicators

        :returns: list of UUIDs representing the indicators
        """
        return list(self._indicators)

    def get_indicator_state(self, identity):
        """Get indicator state

        :param identity: indicator identity

        :returns: indicator state as one of *Lit*, *Blinking*,
                *Off*
        """
        if identity not in self._indicators:
            self._indicators[identity] = 'Lit'

        return self._indicators[identity]

    def set_indicator_state(self, identity, state):
        """Set indicator state

        :param identity: indicator identity
        :param state: indicator state as one of *Lit*, *Blinking*, *Off*

        :raises: `error.FishyError` if desired state can't be set
        """
        if state not in ('Lit', 'Off', 'Blinking'):
            raise error.FishyError(
                'Unknown indicator state %s, ID %s' % (state, identity))

        self._indicators[identity] = state



# Copyright 2019 Red Hat, Inc.
# All Rights Reserved.
#
#    Licensed under the Apache License, Version 2.0 (the "License"); you may
#    not use this file except in compliance with the License. You may obtain
#    a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#    WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#    License for the specific language governing permissions and limitations
#    under the License.

from unittest import mock

from oslotest import base

from sushy_tools.emulator.resources import managers
from sushy_tools import error


class FakeDriverTestCase(base.BaseTestCase):

    def setUp(self):
        super(FakeDriverTestCase, self).setUp()
        self.identity = 'xxx'
        self.systems = mock.Mock(systems=[self.identity])
        self.systems.uuid.return_value = 'xxx'
        self.systems.name.return_value = 'name'
        self.manager = {'UUID': self.identity,
                        'Id': self.identity,
                        'Name': 'name-Manager'}
        self.chassis = mock.Mock(chassis=[])
        self.test_driver = managers.FakeDriver({}, mock.Mock(),
                                               self.systems, self.chassis)

    def test_get_manager_not_found(self):
        self.systems.uuid.side_effect = error.FishyError('boom')
        self.assertRaises(
            error.FishyError, self.test_driver.get_manager, 'foo')

    def test_get_manager_by_uuid(self):
        manager = self.test_driver.get_manager('xxx')
        self.assertEqual(self.manager, manager)

    def test_managers(self):
        result = self.test_driver.managers
        self.assertEqual([self.identity], result)

    def test_managed_systems(self):
        self.assertEqual(
            ['xxx'], self.test_driver.get_managed_systems(self.manager))



#    Licensed under the Apache License, Version 2.0 (the "License"); you may
#    not use this file except in compliance with the License. You may obtain
#    a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#    WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#    License for the specific language governing permissions and limitations
#    under the License.

from unittest import mock

from sushy_tools.emulator import api_utils
from sushy_tools.emulator import main
from sushy_tools.tests.unit.emulator import test_main


class InstanceDeniedTestCase(test_main.EmulatorTestCase):

    def setUp(self):
        super().setUp()
        ctx = main.app.app_context().__enter__()
        self.addCleanup(lambda: ctx.__exit__(None, None, None))

    @mock.patch.dict(main.app.config, {}, clear=True)
    def test_instance_denied_allow_all(self):
        self.assertFalse(api_utils.instance_denied(identity='x'))

    @mock.patch.dict(
        main.app.config, {'SUSHY_EMULATOR_ALLOWED_INSTANCES': {}})
    def test_instance_denied_disallow_all(self):
        self.assertTrue(api_utils.instance_denied(identity='a'))

    def test_instance_denied_undefined_option(self):
        with mock.patch.dict(main.app.config):
            main.app.config.pop('SUSHY_EMULATOR_ALLOWED_INSTANCES', None)
            self.assertFalse(api_utils.instance_denied(identity='a'))

    @mock.patch.dict(
        main.app.config, {'SUSHY_EMULATOR_ALLOWED_INSTANCES': {'a'}})
    def test_instance_denied_allow_some(self):
        self.assertFalse(api_utils.instance_denied(identity='a'))

    @mock.patch.dict(
        main.app.config, {'SUSHY_EMULATOR_ALLOWED_INSTANCES': {'a'}})
    def test_instance_denied_disallow_some(self):
        self.assertTrue(api_utils.instance_denied(identity='b'))



#    Licensed under the Apache License, Version 2.0 (the "License"); you may
#    not use this file except in compliance with the License. You may obtain
#    a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#    WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#    License for the specific language governing permissions and limitations
#    under the License.

import re

import flask

from sushy_tools.emulator import api_utils
from sushy_tools import error


certificate_service = flask.Blueprint(
    'CertificateService', __name__,
    url_prefix='/redfish/v1/CertificateService')


_VMEDIA_URI_PATTERN = re.compile(
    '/redfish/v1/Managers/([^/]+)/VirtualMedia/([^/]+)/Certificates/([^/]+)/?$'
)


@certificate_service.route('', methods=['GET'])
@api_utils.returns_json
def certificate_service_resource():
    api_utils.debug('Serving certificate service')
    return flask.render_template('certificate_service.json')


@certificate_service.route('/CertificateLocations', methods=['GET'])
@api_utils.returns_json
def certificate_service_locations():
    api_utils.debug('Serving certificate locations')
    locations = []
    for mgr in flask.current_app.managers.managers:
        for dev in flask.current_app.vmedia.devices:
            try:
                certs = flask.current_app.vmedia.list_certificates(mgr, dev)
            except error.NotFound:
                api_utils.debug('No certificates for manager %s, virtual '
                                'media device %s', mgr, dev)
                continue

            for cert in certs:
                locations.append(
                    f'/redfish/v1/Managers/{mgr}/VirtualMedia/{dev}'
                    f'/Certificates/{cert.id}'
                )

    return flask.render_template('certificate_locations.json',
                                 certificates=locations)


@certificate_service.route('/Actions/CertificateService.ReplaceCertificate',
                           methods=['POST'])
@api_utils.ensure_instance_access
@api_utils.returns_json
def certificate_service_replace_certificate():
    if not flask.request.json:
        raise error.BadRequest("Empty or malformed certificate")

    try:
        cert_string = flask.request.json['CertificateString']
        cert_type = flask.request.json['CertificateType']
        cert_uri = flask.request.json['CertificateUri']
    except KeyError as exc:
        raise error.BadRequest(f"Missing required parameter {exc}")

    match = _VMEDIA_URI_PATTERN.search(cert_uri)
    if not match:
        raise error.NotFound(
            f"Certificates at URI {cert_uri} are not supported")

    if cert_type != 'PEM':
        raise error.BadRequest(
            f"Only PEM certificates are supported, got {cert_type}")

    manager_id, device, cert_id = match.groups()

    flask.current_app.managers.get_manager(manager_id)
    flask.current_app.vmedia.replace_certificate(
        manager_id, device, cert_id, cert_string, cert_type)

    return '', 204



#    Licensed under the Apache License, Version 2.0 (the "License"); you may
#    not use this file except in compliance with the License. You may obtain
#    a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#    WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#    License for the specific language governing permissions and limitations
#    under the License.

from sushy_tools.emulator.resources import vmedia
from sushy_tools import error
from sushy_tools.tests.unit.emulator import test_main


@test_main.patch_resource('vmedia')
@test_main.patch_resource('managers')
class CertificateServiceTestCase(test_main.EmulatorTestCase):

    def test_root(self, managers_mock, vmedia_mock):
        response = self.app.get('redfish/v1/CertificateService')

        self.assertEqual(200, response.status_code)
        self.assertIn('#CertificateService.ReplaceCertificate',
                      response.json['Actions'])

    def test_replace_ok(self, managers_mock, vmedia_mock):
        response = self.app.post(
            'redfish/v1/CertificateService/Actions/'
            'CertificateService.ReplaceCertificate',
            json={'CertificateString': 'abcd',
                  'CertificateType': 'PEM',
                  'CertificateUri': ('https://host/redfish/v1/Managers/1234'
                                     '/VirtualMedia/CD/Certificates/1')})

        self.assertEqual(204, response.status_code)
        managers_mock.return_value.get_manager.assert_called_once_with('1234')
        vmedia_mock.return_value.replace_certificate.assert_called_once_with(
            '1234', 'CD', '1', 'abcd', 'PEM')

    def test_replace_manager_not_found(self, managers_mock, vmedia_mock):
        managers_mock.return_value.get_manager.side_effect = error.NotFound

        response = self.app.post(
            'redfish/v1/CertificateService/Actions/'
            'CertificateService.ReplaceCertificate',
            json={'CertificateString': 'abcd',
                  'CertificateType': 'PEM',
                  'CertificateUri': ('https://host/redfish/v1/Managers/1234'
                                     '/VirtualMedia/CD/Certificates/1')})

        self.assertEqual(404, response.status_code)
        managers_mock.return_value.get_manager.assert_called_once_with('1234')
        vmedia_mock.return_value.replace_certificate.assert_not_called()

    def test_replace_wrong_uri(self, managers_mock, vmedia_mock):
        response = self.app.post(
            'redfish/v1/CertificateService/Actions/'
            'CertificateService.ReplaceCertificate',
            json={'CertificateString': 'abcd',
                  'CertificateType': 'PEM',
                  'CertificateUri': ('https://host/redfish/v1/Managers/1234'
                                     '/NetworkProtocol/HTTPS/Certificates/1')})

        self.assertEqual(404, response.status_code)
        managers_mock.return_value.get_manager.assert_not_called()
        vmedia_mock.return_value.replace_certificate.assert_not_called()

    def test_replace_missing_string(self, managers_mock, vmedia_mock):
        response = self.app.post(
            'redfish/v1/CertificateService/Actions/'
            'CertificateService.ReplaceCertificate',
            json={'CertificateType': 'PEM',
                  'CertificateUri': ('https://host/redfish/v1/Managers/1234'
                                     '/VirtualMedia/CD/Certificates/1')})

        self.assertEqual(400, response.status_code)
        managers_mock.return_value.get_manager.assert_not_called()
        vmedia_mock.return_value.replace_certificate.assert_not_called()

    def test_replace_wrong_type(self, managers_mock, vmedia_mock):
        response = self.app.post(
            'redfish/v1/CertificateService/Actions/'
            'CertificateService.ReplaceCertificate',
            json={'CertificateString': 'abcd',
                  'CertificateType': 'non-PEM',
                  'CertificateUri': ('https://host/redfish/v1/Managers/1234'
                                     '/VirtualMedia/CD/Certificates/1')})

        self.assertEqual(400, response.status_code)
        managers_mock.return_value.get_manager.assert_not_called()
        vmedia_mock.return_value.replace_certificate.assert_not_called()

    def test_replace_missing_uri(self, managers_mock, vmedia_mock):
        response = self.app.post(
            'redfish/v1/CertificateService/Actions/'
            'CertificateService.ReplaceCertificate',
            json={'CertificateString': 'abcd',
                  'CertificateType': 'PEM'})

        self.assertEqual(400, response.status_code)
        managers_mock.return_value.get_manager.assert_not_called()
        vmedia_mock.return_value.replace_certificate.assert_not_called()

    def test_locations(self, managers_mock, vmedia_mock):
        managers_mock.return_value.managers = ["1", "2"]
        vmedia_mock.return_value.devices = ["CD", "DVD"]
        vmedia_mock.return_value.list_certificates.side_effect = [
            error.NotFound(),
            [vmedia.Certificate("cert1", "abcd", "PEM")],
            [vmedia.Certificate("cert2", "abcd", "PEM")],
            error.NotFound(),
        ]
        response = self.app.get(
            'redfish/v1/CertificateService/CertificateLocations')

        self.assertEqual(200, response.status_code)
        self.assertEqual(
            ['/redfish/v1/Managers/1/VirtualMedia/DVD/Certificates/cert1',
             '/redfish/v1/Managers/2/VirtualMedia/CD/Certificates/cert2'],
            [item['@odata.id']
             for item in response.json['Links']['Certificates']])



# Copyright 2019 Red Hat, Inc.
# All Rights Reserved.
#
#    Licensed under the Apache License, Version 2.0 (the "License"); you may
#    not use this file except in compliance with the License. You may obtain
#    a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#    WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#    License for the specific language governing permissions and limitations
#    under the License.

from unittest import mock

from oslotest import base

from sushy_tools.emulator.resources.drives import StaticDriver


class StaticDriverTestCase(base.BaseTestCase):
    SYSTEM_UUID = "da69abcc-dae0-4913-9a7b-d344043097c0"
    STORAGE_ID = "1"
    DRIVE_COL = [
        {
            "Id": "32ADF365C6C1B7BD",
            "Name": "Drive Sample",
            "CapacityBytes": 899527000000,
            "Protocol": "SAS"
        },
        {
            "Id": "58CFF987G8J2V9KL",
            "Name": "Drive2",
            "CapacityBytes": 12345670000,
            "Protocol": "SATA"
        }
    ]

    CONFIG = {
        'SUSHY_EMULATOR_DRIVES': {
            (SYSTEM_UUID, STORAGE_ID): DRIVE_COL
        }
    }

    def test_get_drives(self):
        test_driver = StaticDriver(self.CONFIG, mock.MagicMock())
        drv_col = test_driver.get_drives(self.SYSTEM_UUID, self.STORAGE_ID)
        self.assertEqual(self.DRIVE_COL, drv_col)

    def test_get_all_drives(self):
        test_driver = StaticDriver(self.CONFIG, mock.MagicMock())
        drives = test_driver.get_all_drives()
        self.assertEqual({('da69abcc-dae0-4913-9a7b-d344043097c0', '1',
                           '32ADF365C6C1B7BD'),
                          ('da69abcc-dae0-4913-9a7b-d344043097c0', '1',
                           '58CFF987G8J2V9KL')}, set(drives))



#    Licensed under the Apache License, Version 2.0 (the "License"); you may
#    not use this file except in compliance with the License. You may obtain
#    a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#    WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#    License for the specific language governing permissions and limitations
#    under the License.

from sushy_tools.emulator.resources import vmedia
from sushy_tools import error
from sushy_tools.tests.unit.emulator import test_main


@test_main.patch_resource('vmedia')
@test_main.patch_resource('managers')
class VirtualMediaTestCase(test_main.EmulatorTestCase):

    def test_virtual_media_collection(self, managers_mock, vmedia_mock):
        managers_mock = managers_mock.return_value
        managers_mock.managers = [self.uuid]
        managers_mock.get_manager.return_value = {'UUID': self.uuid}
        vmedia_mock.return_value.devices = ['CD', 'Floppy']

        response = self.app.get(
            '/redfish/v1/Managers/%s/VirtualMedia' % self.uuid)

        self.assertEqual(200, response.status_code)
        self.assertEqual('Virtual Media Services', response.json['Name'])
        self.assertEqual(2, response.json['Members@odata.count'])
        self.assertEqual(
            ['/redfish/v1/Managers/%s/VirtualMedia/CD' % self.uuid,
             '/redfish/v1/Managers/%s/VirtualMedia/Floppy' % self.uuid],
            [m['@odata.id'] for m in response.json['Members']])

    def test_virtual_media_collection_empty(self, managers_mock, vmedia_mock):
        vmedia_mock.return_value.get_devices.return_value = []

        response = self.app.get(
            'redfish/v1/Managers/' + self.uuid + '/VirtualMedia')

        self.assertEqual(200, response.status_code)
        self.assertEqual('Virtual Media Services', response.json['Name'])
        self.assertEqual(0, response.json['Members@odata.count'])
        self.assertEqual([], response.json['Members'])

    def test_virtual_media(self, managers_mock, vmedia_mock):
        vmedia_mock = vmedia_mock.return_value
        vmedia_mock.get_device_name.return_value = 'CD'
        vmedia_mock.get_device_media_types.return_value = [
            'CD', 'DVD']
        vmedia_mock.get_device_image_info.return_value = vmedia.DeviceInfo(
            'image-of-a-fish', 'fishy.iso', True, True, '', '', False)

        response = self.app.get(
            '/redfish/v1/Managers/%s/VirtualMedia/CD' % self.uuid)

        self.assertEqual(200, response.status_code, response.json)
        self.assertEqual('CD', response.json['Id'])
        self.assertEqual(['CD', 'DVD'], response.json['MediaTypes'])
        self.assertEqual('fishy.iso', response.json['Image'])
        self.assertEqual('image-of-a-fish', response.json['ImageName'])
        self.assertTrue(response.json['Inserted'])
        self.assertTrue(response.json['WriteProtected'])
        self.assertEqual('', response.json['UserName'])
        self.assertEqual('', response.json['Password'])
        self.assertFalse(response.json['VerifyCertificate'])
        self.assertEqual(
            '/redfish/v1/Managers/%s/VirtualMedia/CD/Certificates' % self.uuid,
            response.json['Certificates']['@odata.id'])

    def test_virtual_media_with_auth(self, managers_mock, vmedia_mock):
        vmedia_mock = vmedia_mock.return_value
        vmedia_mock.get_device_name.return_value = 'CD'
        vmedia_mock.get_device_media_types.return_value = [
            'CD', 'DVD']
        vmedia_mock.get_device_image_info.return_value = vmedia.DeviceInfo(
            'image-of-a-fish', 'fishy.iso', True, True, 'Admin', 'Secret',
            False)

        response = self.app.get(
            '/redfish/v1/Managers/%s/VirtualMedia/CD' % self.uuid)

        self.assertEqual(200, response.status_code, response.json)
        self.assertEqual('CD', response.json['Id'])
        self.assertEqual(['CD', 'DVD'], response.json['MediaTypes'])
        self.assertEqual('fishy.iso', response.json['Image'])
        self.assertEqual('image-of-a-fish', response.json['ImageName'])
        self.assertTrue(response.json['Inserted'])
        self.assertTrue(response.json['WriteProtected'])
        self.assertEqual('Admin', response.json['UserName'])
        self.assertEqual('******', response.json['Password'])
        self.assertFalse(response.json['VerifyCertificate'])

    def test_virtual_media_not_found(self, managers_mock, vmedia_mock):
        vmedia_mock.return_value.get_device_name.side_effect = error.NotFound

        response = self.app.get(
            '/redfish/v1/Managers/%s/VirtualMedia/DVD-ROM' % self.uuid)

        self.assertEqual(404, response.status_code)

    def test_virtual_media_update(self, managers_mock, vmedia_mock):
        response = self.app.patch(
            '/redfish/v1/Managers/%s/VirtualMedia/CD' % self.uuid,
            json={'VerifyCertificate': True})

        self.assertEqual(204, response.status_code)
        vmedia_mock = vmedia_mock.return_value
        vmedia_mock.update_device_info.assert_called_once_with(
            self.uuid, 'CD', verify=True)

    def test_virtual_media_update_not_found(self, managers_mock, vmedia_mock):
        vmedia_mock = vmedia_mock.return_value
        vmedia_mock.update_device_info.side_effect = error.NotFound

        response = self.app.patch(
            '/redfish/v1/Managers/%s/VirtualMedia/DVD-ROM' % self.uuid,
            json={'VerifyCertificate': True})

        self.assertEqual(404, response.status_code)

    def test_virtual_media_update_invalid(self, managers_mock, vmedia_mock):
        response = self.app.patch(
            '/redfish/v1/Managers/%s/VirtualMedia/CD' % self.uuid,
            json={'VerifyCertificate': 'banana'})

        self.assertEqual(400, response.status_code)

    def test_virtual_media_update_empty(self, managers_mock, vmedia_mock):
        response = self.app.patch(
            '/redfish/v1/Managers/%s/VirtualMedia/CD' % self.uuid)

        self.assertEqual(400, response.status_code)

    def test_virtual_media_insert(self, managers_mock, vmedia_mock):
        response = self.app.post(
            '/redfish/v1/Managers/%s/VirtualMedia/CD/Actions/'
            'VirtualMedia.InsertMedia' % self.uuid,
            json={"Image": "http://fish.iso"})

        self.assertEqual(204, response.status_code)

        vmedia_mock.return_value.insert_image.called_once_with(
            'CD', 'http://fish.iso', True, False)

    def test_virtual_media_eject(self, managers_mock, vmedia_mock):
        response = self.app.post(
            '/redfish/v1/Managers/%s/VirtualMedia/CD/Actions/'
            'VirtualMedia.EjectMedia' % self.uuid,
            json={})

        self.assertEqual(204, response.status_code)

        vmedia_mock.return_value.eject_image.called_once_with('CD')

    def test_virtual_media_certificates(self, managers_mock, vmedia_mock):
        vmedia_mock.return_value.list_certificates.return_value = [
            vmedia.Certificate('1', 'PEM', 'abcd'),
            vmedia.Certificate('2', 'PEM', 'dcba'),
        ]
        response = self.app.get(
            f'/redfish/v1/Managers/{self.uuid}/VirtualMedia/CD/Certificates')

        self.assertEqual(200, response.status_code, response.json)
        self.assertEqual(2, response.json['Members@odata.count'])
        for index, member in enumerate(response.json['Members']):
            self.assertTrue(member['@odata.id'].endswith(
                f'/redfish/v1/Managers/{self.uuid}/VirtualMedia/CD'
                f'/Certificates/{index+1}'), member['@odata.id'])
        self.assertEqual(['PEM'],
                         response.json['@Redfish.SupportedCertificates'])

    def test_virtual_media_certificates_manager_not_found(self, managers_mock,
                                                          vmedia_mock):
        managers_mock.return_value.get_manager.side_effect = error.NotFound
        response = self.app.get(
            f'/redfish/v1/Managers/{self.uuid}/VirtualMedia/CD/Certificates')

        self.assertEqual(404, response.status_code, response.json)

    def test_virtual_media_add_certificate(self, managers_mock, vmedia_mock):
        vmedia_mock.return_value.add_certificate.return_value = \
            vmedia.Certificate('9', 'abcd', 'PEM')

        response = self.app.post(
            f'/redfish/v1/Managers/{self.uuid}/VirtualMedia/CD/Certificates',
            json={'CertificateString': 'abcd', 'CertificateType': 'PEM'})

        self.assertEqual(204, response.status_code, response.data)
        self.assertIn(
            f'/redfish/v1/Managers/{self.uuid}/VirtualMedia/CD/Certificates/9',
            response.headers['Location'])

    def test_virtual_media_add_certificate_no_string(self, managers_mock,
                                                     vmedia_mock):
        response = self.app.post(
            f'/redfish/v1/Managers/{self.uuid}/VirtualMedia/CD/Certificates',
            json={'CertificateType': 'PEM'})

        self.assertEqual(400, response.status_code, response.data)

    def test_virtual_media_add_certificate_bad_type(self, managers_mock,
                                                    vmedia_mock):
        response = self.app.post(
            f'/redfish/v1/Managers/{self.uuid}/VirtualMedia/CD/Certificates',
            json={'CertificateString': 'abcd', 'CertificateType': 'non-PEM'})

        self.assertEqual(400, response.status_code, response.data)

    def test_virtual_media_get_certificate(self, managers_mock, vmedia_mock):
        vmedia_mock.return_value.list_certificates.return_value = [
            vmedia.Certificate('1', 'abcd', 'PEM'),
            vmedia.Certificate('2', 'dcba', 'PEM'),
        ]
        response = self.app.get(
            f'/redfish/v1/Managers/{self.uuid}/VirtualMedia/CD/Certificates/2')

        self.assertEqual(200, response.status_code, response.json)
        self.assertIn('Id', response.json)
        self.assertEqual('2', response.json['Id'])
        self.assertEqual('dcba', response.json['CertificateString'])
        self.assertEqual('PEM', response.json['CertificateType'])

    def test_virtual_media_get_certificate_manager_not_found(self,
                                                             managers_mock,
                                                             vmedia_mock):
        managers_mock.return_value.get_manager.side_effect = error.NotFound
        response = self.app.get(
            f'/redfish/v1/Managers/{self.uuid}/VirtualMedia/CD/Certificates/2')

        self.assertEqual(404, response.status_code, response.json)

    def test_virtual_media_delete_certificate(self, managers_mock,
                                              vmedia_mock):
        response = self.app.delete(
            f'/redfish/v1/Managers/{self.uuid}/VirtualMedia/CD/Certificates/2')

        self.assertEqual(204, response.status_code, response.data)
        vmedia_mock.return_value.delete_certificate.assert_called_once_with(
            self.uuid, 'CD', '2')

    def test_virtual_media_delete_certificate_manager_not_found(self,
                                                                managers_mock,
                                                                vmedia_mock):
        managers_mock.return_value.get_manager.side_effect = error.NotFound
        response = self.app.delete(
            f'/redfish/v1/Managers/{self.uuid}/VirtualMedia/CD/Certificates/2')

        self.assertEqual(404, response.status_code, response.json)
        vmedia_mock.return_value.delete_certificate.assert_not_called()



#    Licensed under the Apache License, Version 2.0 (the "License"); you may
#    not use this file except in compliance with the License. You may obtain
#    a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#    WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#    License for the specific language governing permissions and limitations
#    under the License.

import copy
import random
import time

from sushy_tools.emulator import memoize
from sushy_tools.emulator.resources.systems.base import AbstractSystemsDriver
from sushy_tools import error


DEFAULT_UUID = '27946b59-9e44-4fa7-8e91-f3527a1ef094'


class FakeDriver(AbstractSystemsDriver):
    """Fake driver"""

    @classmethod
    def initialize(cls, config, logger):
        config.setdefault('SUSHY_EMULATOR_FAKE_SYSTEMS', [
            {
                'uuid': DEFAULT_UUID,
                'name': 'fake',
                'power_state': 'Off',
            }
        ])
        cls._config = config
        cls._logger = logger
        return cls

    def __init__(self):
        super().__init__()
        self._systems = memoize.PersistentDict()
        if hasattr(self._systems, 'make_permanent'):
            self._systems.make_permanent(
                self._config.get('SUSHY_EMULATOR_STATE_DIR'), 'fakedriver')

        for system in self._config['SUSHY_EMULATOR_FAKE_SYSTEMS']:
            # Be careful to reduce racing with other processes
            if system['uuid'] not in self._systems:
                self._systems[system['uuid']] = copy.deepcopy(system)

        self._by_name = {
            system['name']: uuid
            for uuid, system in self._systems.items()
        }

    def _update_if_needed(self, system):
        pending_power = system.get('pending_power')
        if pending_power and time.time() >= pending_power['apply_time']:
            self._update(system,
                         power_state=pending_power['power_state'],
                         pending_power=None)

        return system

    def _get(self, identity):
        try:
            result = self._systems[identity]
        except KeyError:
            try:
                uuid = self._by_name[identity]
            except KeyError:
                raise error.NotFound(f'Fake system {identity} was not found')
            else:
                raise error.AliasAccessError(uuid)

        # NOTE(dtantsur): since the state change can only be observed after
        # a _get() call, we can cheat a bit and update it on reading.
        return self._update_if_needed(result)

    def _update(self, system, **changes):
        if isinstance(system, str):
            system = self._get(system)
        system.update(changes)
        self._systems[system['uuid']] = system

    @property
    def driver(self):
        return '<fake>'

    @property
    def systems(self):
        return list(self._systems)

    def uuid(self, identity):
        try:
            return self._get(identity)['uuid']
        except error.AliasAccessError as exc:
            return str(exc)

    def name(self, identity):
        try:
            return self._get(identity)['name']
        except error.AliasAccessError:
            return identity

    def get_power_state(self, identity):
        return self._get(identity)['power_state']

    def set_power_state(self, identity, state):
        # hardware actions are not immediate
        apply_time = int(time.time()) + random.randint(1, 11)

        system = self._get(identity)

        if 'On' in state:
            pending_state = 'On'
        elif state in ('ForceOff', 'GracefulShutdown'):
            pending_state = 'Off'
        elif 'Restart' in state:
            system['power_state'] = 'Off'
            pending_state = 'On'
        else:
            raise error.NotSupportedError(
                f'Power state {state} is not supported')

        if system['power_state'] != pending_state:
            self._update(system, pending_power={
                'power_state': pending_state,
                'apply_time': apply_time,
            })

    def get_boot_device(self, identity):
        return self._get(identity).get('boot_device', 'Hdd')

    def set_boot_device(self, identity, boot_source):
        self._update(identity, boot_device=boot_source)

    def get_boot_mode(self, identity):
        return self._get(identity).get('boot_mode', 'UEFI')

    def set_boot_mode(self, identity, boot_mode):
        self._update(identity, boot_mode=boot_mode)

    def get_boot_image(self, identity, device):
        devinfo = self._get(identity).get('boot_image') or {}
        return devinfo.get(device) or (None, False, False)

    def set_boot_image(self, identity, device, boot_image=None,
                       write_protected=True):
        system = self._get(identity)
        devinfo = system.get('boot_image') or {}
        devinfo[device] = (boot_image, write_protected, bool(boot_image))
        self._update(system, boot_image=devinfo)



#    Licensed under the Apache License, Version 2.0 (the "License"); you may
#    not use this file except in compliance with the License. You may obtain
#    a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#    WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#    License for the specific language governing permissions and limitations
#    under the License.

import functools

import flask


def debug(*args, **kwargs):
    flask.current_app.logger.debug(*args, **kwargs)


def info(*args, **kwargs):
    flask.current_app.logger.info(*args, **kwargs)


def warning(*args, **kwargs):
    flask.current_app.logger.warning(*args, **kwargs)


def error(*args, **kwargs):
    flask.current_app.logger.error(*args, **kwargs)


def instance_denied(**kwargs):
    deny = True

    try:
        deny = (kwargs['identity'] not in
                flask.current_app.config['SUSHY_EMULATOR_ALLOWED_INSTANCES'])

    except KeyError:
        deny = False

    finally:
        if deny:
            warning('Instance %s access denied', kwargs.get('identity'))

        return deny


def ensure_instance_access(decorated_func):
    @functools.wraps(decorated_func)
    def decorator(*args, **kwargs):
        if instance_denied(**kwargs):
            raise error.NotFound()

        return decorated_func(*args, **kwargs)

    return decorator


def returns_json(decorated_func):
    @functools.wraps(decorated_func)
    def decorator(*args, **kwargs):
        response = decorated_func(*args, **kwargs)
        if isinstance(response, flask.Response):
            return response
        if isinstance(response, tuple):
            contents, status, *headers = response
        else:
            contents, status, headers = response, 200, ()
        kwargs = {'headers': headers[0]} if headers else {}
        return flask.Response(response=contents, status=status,
                              content_type='application/json', **kwargs)

    return decorator



# Copyright 2018 Red Hat, Inc.
# All Rights Reserved.
#
#    Licensed under the Apache License, Version 2.0 (the "License"); you may
#    not use this file except in compliance with the License. You may obtain
#    a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#    WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#    License for the specific language governing permissions and limitations
#    under the License.

import math

from sushy_tools.emulator import memoize
from sushy_tools.emulator.resources.systems.base import AbstractSystemsDriver
from sushy_tools import error

try:
    import openstack

except ImportError:
    openstack = None


is_loaded = bool(openstack)


class OpenStackDriver(AbstractSystemsDriver):
    """OpenStack driver"""

    NOVA_POWER_STATE_ON = 1

    BOOT_DEVICE_MAP = {
        'Pxe': 'network',
        'Hdd': 'hd',
        'Cd': 'cdrom',
    }

    BOOT_DEVICE_MAP_REV = {v: k for k, v in BOOT_DEVICE_MAP.items()}

    BOOT_MODE_MAP = {
        'Legacy': 'bios',
        'UEFI': 'uefi',
    }

    BOOT_MODE_MAP_REV = {v: k for k, v in BOOT_MODE_MAP.items()}

    PERMANENT_CACHE = {}

    @classmethod
    def initialize(cls, config, logger, os_cloud, *args, **kwargs):
        cls._config = config
        cls._logger = logger
        cls._os_cloud = os_cloud

        cls._cc = openstack.connect(cloud=os_cloud)

        return cls

    @memoize.memoize()
    def _get_instance(self, identity):
        instance = self._cc.get_server(identity)
        if instance:
            if identity != instance.id:
                raise error.AliasAccessError(instance.id)

            return instance

        msg = ('Error finding instance by UUID "%(identity)s" at OS '
               'cloud %(os_cloud)s"' % {'identity': identity,
                                        'os_cloud': self._os_cloud})

        self._logger.debug(msg)

        raise error.NotFound(msg)

    @memoize.memoize(permanent_cache=PERMANENT_CACHE)
    def _get_flavor(self, identity):
        instance = self._get_instance(identity)
        return self._cc.get_flavor(instance.flavor.original_name)

    @memoize.memoize(permanent_cache=PERMANENT_CACHE)
    def _get_image_info(self, identity):
        return self._cc.image.find_image(identity)

    def _get_server_metadata(self, identity):
        return self._cc.compute.get_server_metadata(identity).to_dict()

    def _set_server_metadata(self, identity, metadata):
        self._cc.compute.set_server_metadata(identity, metadata)

    @property
    def driver(self):
        """Return human-friendly driver description

        :returns: driver description as `str`
        """
        return '<OpenStack compute>'

    @property
    def systems(self):
        """Return available computer systems

        :returns: list of UUIDs representing the systems
        """
        return [server.id for server in self._cc.list_servers()]

    def uuid(self, identity):
        """Get computer system UUID by name

        :param identity: OpenStack instance name or ID

        :returns: computer system UUID
        """
        instance = self._get_instance(identity)
        return instance.id

    def name(self, identity):
        """Get computer system name by name

        :param identity: OpenStack instance name or ID

        :returns: computer system name
        """
        instance = self._get_instance(identity)
        return instance.name

    def get_power_state(self, identity):
        """Get computer system power state

        :param identity: OpenStack instance name or ID

        :returns: *On* or *Off*`str` or `None`
            if power state can't be determined
        """
        try:
            instance = self._get_instance(identity)

        except error.FishyError:
            return

        if instance.power_state == self.NOVA_POWER_STATE_ON:
            return 'On'

        return 'Off'

    def set_power_state(self, identity, state):
        """Set computer system power state

        :param identity: OpenStack instance name or ID
        :param state: string literal requesting power state transition.
            Valid values  are: *On*, *ForceOn*, *ForceOff*, *GracefulShutdown*,
            *GracefulRestart*, *ForceRestart*, *Nmi*.

        :raises: `error.FishyError` if power state can't be set

        """
        instance = self._get_instance(identity)

        if state in ('On', 'ForceOn'):
            if instance.power_state != self.NOVA_POWER_STATE_ON:
                self._cc.compute.start_server(instance.id)

        elif state == 'ForceOff':
            if instance.power_state == self.NOVA_POWER_STATE_ON:
                self._cc.compute.stop_server(instance.id)

        elif state == 'GracefulShutdown':
            if instance.power_state == self.NOVA_POWER_STATE_ON:
                self._cc.compute.stop_server(instance.id)

        elif state == 'GracefulRestart':
            if instance.power_state == self.NOVA_POWER_STATE_ON:
                self._cc.compute.reboot_server(
                    instance.id, reboot_type='SOFT'
                )

        elif state == 'ForceRestart':
            if instance.power_state == self.NOVA_POWER_STATE_ON:
                self._cc.compute.reboot_server(
                    instance.id, reboot_type='HARD'
                )

        # NOTE(etingof) can't support `state == "Nmi"` as
        # openstacksdk does not seem to support that
        else:
            raise error.BadRequest(
                'Unknown ResetType "%(state)s"' % {'state': state})

    def get_boot_device(self, identity):
        """Get computer system boot device name

        :param identity: OpenStack instance name or ID

        :returns: boot device name as `str` or `None` if device name
            can't be determined. Valid values are: *Pxe*, *Hdd*, *Cd*.
        """
        try:
            instance = self._get_instance(identity)

        except error.FishyError:
            return

        metadata = self._get_server_metadata(instance.id)

        # NOTE(etingof): the following probably only works with
        # libvirt-backed compute nodes

        if metadata.get('libvirt:pxe-first'):
            return self.BOOT_DEVICE_MAP_REV['network']

        else:
            return self.BOOT_DEVICE_MAP_REV['hd']

    def set_boot_device(self, identity, boot_source):
        """Set computer system boot device name

        :param identity: OpenStack instance name or ID
        :param boot_source: string literal requesting boot device
            change on the system. Valid values are: *Pxe*, *Hdd*, *Cd*.

        :raises: `error.FishyError` if boot device can't be set
        """
        instance = self._get_instance(identity)

        try:
            target = self.BOOT_DEVICE_MAP[boot_source]

        except KeyError:
            msg = ('Unknown power state requested: '
                   '%(boot_source)s' % {'boot_source': boot_source})

            raise error.BadRequest(msg)

        # NOTE(etingof): the following probably only works with
        # libvirt-backed compute nodes
        self._cc.compute.set_server_metadata(
            instance.id, **{'libvirt:pxe-first': '1'
                            if target == 'network' else ''}
        )

    def get_boot_mode(self, identity):
        """Get computer system boot mode.

        :returns: either *UEFI* or *Legacy* as `str` or `None` if
            current boot mode can't be determined
        """
        instance = self._get_instance(identity)

        image = self._get_image_info(instance.image['id'])

        hw_firmware_type = getattr(image, 'hw_firmware_type', None)

        return self.BOOT_MODE_MAP_REV.get(hw_firmware_type)

    def set_boot_mode(self, identity, boot_mode):
        """Set computer system boot mode.

        :param boot_mode: string literal requesting boot mode
            change on the system. Valid values are: *UEFI*, *Legacy*.

        :raises: `error.FishyError` if boot mode can't be set
        """
        # just to make sure passed identity exists
        self._get_instance(identity)

        msg = ('The cloud driver %(driver)s does not allow changing boot '
               'mode through Redfish' % {'driver': self.driver})

        raise error.NotSupportedError(msg)

    def get_total_memory(self, identity):
        """Get computer system total memory

        :param identity: OpenStack instance name or ID

        :returns: available RAM in GiB as `int` or `None` if total memory
            count can't be determined
        """
        try:
            flavor = self._get_flavor(identity)

        except error.FishyError:
            return

        return int(math.ceil(flavor.ram / 1024.))

    def get_total_cpus(self, identity):
        """Get computer system total count of available CPUs

        :param identity: OpenStack instance name or ID

        :returns: available CPU count as `int` or `None`
            if total memory count can't be determined
        """
        try:
            flavor = self._get_flavor(identity)

        except error.FishyError:
            return

        return flavor.vcpus

    def get_nics(self, identity):
        """Get server's network interfaces

        Use MAC address as network interface's id

        :param identity: OpenStack instance name or ID

        :returns: list of dictionaries with NIC attributes (id and mac)
        """
        instance = self._get_instance(identity)
        macs = set()
        if not instance.addresses:
            return macs

        for addresses in instance.addresses.values():
            for adr in addresses:
                try:
                    macs.add(adr['OS-EXT-IPS-MAC:mac_addr'])
                except KeyError:
                    self._logger.warning(
                        'Could not find MAC address in %s', adr)
        return [{'id': mac, 'mac': mac}
                for mac in macs]



# Copyright 2017 Red Hat, Inc.
# All Rights Reserved.
#
#    Licensed under the Apache License, Version 2.0 (the "License"); you may
#    not use this file except in compliance with the License. You may obtain
#    a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#    WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#    License for the specific language governing permissions and limitations
#    under the License.

import argparse
from datetime import datetime
import json
import os
import ssl
import sys

import flask
from ironic_lib import auth_basic
from werkzeug import exceptions as wz_exc

from sushy_tools.emulator import api_utils
from sushy_tools.emulator.controllers import certificate_service as certctl
from sushy_tools.emulator.controllers import virtual_media as vmctl
from sushy_tools.emulator import memoize
from sushy_tools.emulator.resources import chassis as chsdriver
from sushy_tools.emulator.resources import drives as drvdriver
from sushy_tools.emulator.resources import indicators as inddriver
from sushy_tools.emulator.resources import managers as mgrdriver
from sushy_tools.emulator.resources import storage as stgdriver
from sushy_tools.emulator.resources.systems import fakedriver
from sushy_tools.emulator.resources.systems import libvirtdriver
from sushy_tools.emulator.resources.systems import novadriver
from sushy_tools.emulator.resources import vmedia as vmddriver
from sushy_tools.emulator.resources import volumes as voldriver
from sushy_tools import error


def _render_error(message):
    return {
        "error": {
            "code": "Base.1.0.GeneralError",
            "message": message,
            "@Message.ExtendedInfo": [
                {
                    "@odata.type": ("/redfish/v1/$metadata"
                                    "#Message.1.0.0.Message"),
                    "MessageId": "Base.1.0.GeneralError"
                }
            ]
        }
    }


class RedfishAuthMiddleware(auth_basic.BasicAuthMiddleware):

    _EXCLUDE_PATHS = frozenset(['', 'redfish', 'redfish/v1'])

    def __call__(self, env, start_response):
        path = env.get('PATH_INFO', '')
        if path.strip('/') in self._EXCLUDE_PATHS:
            return self.app(env, start_response)
        else:
            return super().__call__(env, start_response)

    def format_exception(self, e):
        response = super().format_exception(e)
        response.json_body = _render_error(str(e))
        return response


class Application(flask.Flask):

    def __init__(self):
        super().__init__(__name__)
        # Turn off strict_slashes on all routes
        self.url_map.strict_slashes = False
        # This is needed for WSGI since it cannot process argv
        self.configure(config_file=os.environ.get('SUSHY_EMULATOR_CONFIG'))

        @self.before_request
        def reset_cache():
            self._cache = {}

    def configure(self, config_file=None, extra_config=None):
        if config_file:
            self.config.from_pyfile(config_file)
        if extra_config:
            self.config.update(extra_config)

        auth_file = self.config.get("SUSHY_EMULATOR_AUTH_FILE")
        if auth_file and not isinstance(self.wsgi_app, RedfishAuthMiddleware):
            self.wsgi_app = RedfishAuthMiddleware(self.wsgi_app, auth_file)

    @property
    @memoize.memoize()
    def systems(self):
        fake = self.config.get('SUSHY_EMULATOR_FAKE_DRIVER')
        os_cloud = self.config.get('SUSHY_EMULATOR_OS_CLOUD')

        if fake:
            result = fakedriver.FakeDriver.initialize(
                self.config, self.logger)()

        elif os_cloud:
            if not novadriver.is_loaded:
                self.logger.error('Nova driver not loaded')
                sys.exit(1)

            result = novadriver.OpenStackDriver.initialize(
                self.config, self.logger, os_cloud)()

        else:
            if not libvirtdriver.is_loaded:
                self.logger.error('libvirt driver not loaded')
                sys.exit(1)

            libvirt_uri = self.config.get('SUSHY_EMULATOR_LIBVIRT_URI', '')

            result = libvirtdriver.LibvirtDriver.initialize(
                self.config, self.logger, libvirt_uri)()

        self.logger.debug('Initialized system resource backed by %s driver',
                          result)
        return result

    @property
    @memoize.memoize()
    def managers(self):
        return mgrdriver.FakeDriver(self.config, self.logger,
                                    self.systems, self.chassis)

    @property
    @memoize.memoize()
    def chassis(self):
        return chsdriver.StaticDriver(self.config, self.logger)

    @property
    @memoize.memoize()
    def indicators(self):
        return inddriver.StaticDriver(self.config, self.logger)

    @property
    @memoize.memoize()
    def vmedia(self):
        return vmddriver.StaticDriver(self.config, self.logger)

    @property
    @memoize.memoize()
    def storage(self):
        return stgdriver.StaticDriver(self.config, self.logger)

    @property
    @memoize.memoize()
    def drives(self):
        return drvdriver.StaticDriver(self.config, self.logger)

    @property
    @memoize.memoize()
    def volumes(self):
        return voldriver.StaticDriver(self.config, self.logger)


app = Application()
app.register_blueprint(certctl.certificate_service)
app.register_blueprint(vmctl.virtual_media)


@app.errorhandler(Exception)
@api_utils.returns_json
def all_exception_handler(message):
    if isinstance(message, error.AliasAccessError):
        url = flask.url_for(flask.request.endpoint, identity=message.args[0])
        return flask.redirect(url, code=307, Response=flask.Response)

    code = getattr(message, 'code', 500)
    if (isinstance(message, error.FishyError)
            or isinstance(message, wz_exc.HTTPException)):
        app.logger.debug(
            'Request failed with %s: %s', message.__class__.__name__, message)
    else:
        app.logger.exception(
            'Unexpected %s: %s', message.__class__.__name__, message)

    return flask.render_template('error.json', message=message), code


@app.route('/redfish/v1/')
@api_utils.returns_json
def root_resource():
    return flask.render_template('root.json')


@app.route('/redfish/v1/Chassis')
@api_utils.returns_json
def chassis_collection_resource():
    app.logger.debug('Serving chassis list')

    return flask.render_template(
        'chassis_collection.json',
        manager_count=len(app.chassis.chassis),
        chassis=app.chassis.chassis)


@app.route('/redfish/v1/Chassis/<identity>', methods=['GET', 'PATCH'])
@api_utils.returns_json
def chassis_resource(identity):
    chassis = app.chassis

    uuid = chassis.uuid(identity)

    if flask.request.method == 'GET':

        app.logger.debug('Serving resources for chassis "%s"', identity)

        # the first chassis gets all resources
        if uuid == chassis.chassis[0]:
            systems = app.systems.systems
            managers = app.managers.managers
            storage = app.storage.get_all_storage()
            drives = app.drives.get_all_drives()

        else:
            systems = []
            managers = []
            storage = []
            drives = []

        return flask.render_template(
            'chassis.json',
            identity=identity,
            name=chassis.name(identity),
            uuid=uuid,
            contained_by=None,
            contained_systems=systems,
            contained_managers=managers,
            contained_chassis=[],
            managers=managers[:1],
            indicator_led=app.indicators.get_indicator_state(uuid),
            storage=storage,
            drives=drives
        )

    elif flask.request.method == 'PATCH':
        indicator_led_state = flask.request.json.get('IndicatorLED')
        if not indicator_led_state:
            return 'PATCH only works for IndicatorLED element', 400

        app.indicators.set_indicator_state(uuid, indicator_led_state)

        app.logger.info('Set indicator LED to "%s" for chassis "%s"',
                        indicator_led_state, identity)

        return '', 204


@app.route('/redfish/v1/Chassis/<identity>/Thermal', methods=['GET'])
@api_utils.returns_json
def thermal_resource(identity):
    chassis = app.chassis

    uuid = chassis.uuid(identity)

    app.logger.debug(
        'Serving thermal resources for chassis "%s"', identity)

    # the first chassis gets all resources
    if uuid == chassis.chassis[0]:
        systems = app.systems.systems

    else:
        systems = []

    return flask.render_template(
        'thermal.json',
        chassis=identity,
        systems=systems
    )


@app.route('/redfish/v1/Managers')
@api_utils.returns_json
def manager_collection_resource():
    app.logger.debug('Serving managers list')

    return flask.render_template(
        'manager_collection.json',
        manager_count=len(app.managers.managers),
        managers=app.managers.managers)


def jsonify(obj_type, obj_version, obj):
    obj.update({
        "@odata.type": "#{0}.{1}.{0}".format(obj_type, obj_version),
        "@odata.context": "/redfish/v1/$metadata#{0}.{0}".format(obj_type),
        "@Redfish.Copyright": ("Copyright 2014-2017 Distributed Management "
                               "Task Force, Inc. (DMTF). For the full DMTF "
                               "copyright policy, see http://www.dmtf.org/"
                               "about/policies/copyright.")
    })
    return flask.jsonify(obj)


@app.route('/redfish/v1/Managers/<identity>', methods=['GET'])
@api_utils.returns_json
def manager_resource(identity):
    app.logger.debug('Serving resources for manager "%s"', identity)

    manager = app.managers.get_manager(identity)
    systems = app.managers.get_managed_systems(manager)
    chassis = app.managers.get_managed_chassis(manager)

    uuid = manager['UUID']
    return jsonify('Manager', 'v1_3_1', {
        "Id": manager['Id'],
        "Name": manager.get('Name'),
        "UUID": uuid,
        "ServiceEntryPointUUID": manager.get('ServiceEntryPointUUID'),
        "ManagerType": "BMC",
        "Description": "Contoso BMC",
        "Model": "Joo Janta 200",
        "DateTime": datetime.now().strftime('%Y-%M-%dT%H:%M:%S+00:00'),
        "DateTimeLocalOffset": "+00:00",
        "Status": {
            "State": "Enabled",
            "Health": "OK"
        },
        "PowerState": "On",
        "FirmwareVersion": "1.00",
        "VirtualMedia": {
            "@odata.id": "/redfish/v1/Managers/%s/VirtualMedia" % uuid
        },
        "Links": {
            "ManagerForServers": [
                {
                    "@odata.id": "/redfish/v1/Systems/%s" % system
                }
                for system in systems
            ],
            "ManagerForChassis": [
                {
                    "@odata.id": "/redfish/v1/Chassis/%s" % ch
                }
                for ch in chassis
            ]
        },
        "@odata.id": "/redfish/v1/Managers/%s" % uuid
    })


@app.route('/redfish/v1/Systems')
@api_utils.returns_json
def system_collection_resource():
    systems = [system for system in app.systems.systems
               if not api_utils.instance_denied(identity=system)]

    app.logger.debug('Serving systems list')

    return flask.render_template(
        'system_collection.json', system_count=len(systems), systems=systems)


@app.route('/redfish/v1/Systems/<identity>', methods=['GET', 'PATCH'])
@api_utils.ensure_instance_access
@api_utils.returns_json
def system_resource(identity):
    if flask.request.method == 'GET':

        app.logger.debug('Serving resources for system "%s"', identity)

        def try_get(call):
            try:
                return call(identity)
            except error.NotSupportedError:
                return None

        return flask.render_template(
            'system.json',
            identity=identity,
            name=app.systems.name(identity),
            uuid=app.systems.uuid(identity),
            power_state=app.systems.get_power_state(identity),
            total_memory_gb=try_get(app.systems.get_total_memory),
            total_cpus=try_get(app.systems.get_total_cpus),
            boot_source_target=app.systems.get_boot_device(identity),
            boot_source_mode=try_get(app.systems.get_boot_mode),
            managers=app.managers.get_managers_for_system(identity),
            chassis=app.chassis.chassis[:1],
            indicator_led=app.indicators.get_indicator_state(
                app.systems.uuid(identity))
        )

    elif flask.request.method == 'PATCH':
        boot = flask.request.json.get('Boot')
        indicator_led_state = flask.request.json.get('IndicatorLED')
        if not boot and not indicator_led_state:
            return ('PATCH only works for Boot and '
                    'IndicatorLED elements'), 400

        if boot:
            target = boot.get('BootSourceOverrideTarget')

            if target:
                # NOTE(lucasagomes): In libvirt we always set the boot
                # device frequency to "continuous" so, we are ignoring the
                # BootSourceOverrideEnabled element here

                app.systems.set_boot_device(identity, target)

                app.logger.info('Set boot device to "%s" for system "%s"',
                                target, identity)

            mode = boot.get('BootSourceOverrideMode')

            if mode:
                app.systems.set_boot_mode(identity, mode)

                app.logger.info('Set boot mode to "%s" for system "%s"',
                                mode, identity)

            if not target and not mode:
                return ('Missing the BootSourceOverrideTarget and/or '
                        'BootSourceOverrideMode element', 400)

        if indicator_led_state:
            app.indicators.set_indicator_state(
                app.systems.uuid(identity), indicator_led_state)

            app.logger.info('Set indicator LED to "%s" for system "%s"',
                            indicator_led_state, identity)

        return '', 204


@app.route('/redfish/v1/Systems/<identity>/EthernetInterfaces',
           methods=['GET'])
@api_utils.ensure_instance_access
@api_utils.returns_json
def ethernet_interfaces_collection(identity):
    nics = app.systems.get_nics(identity)

    return flask.render_template(
        'ethernet_interfaces_collection.json', identity=identity,
        nics=nics)


@app.route('/redfish/v1/Systems/<identity>/EthernetInterfaces/<nic_id>',
           methods=['GET'])
@api_utils.ensure_instance_access
@api_utils.returns_json
def ethernet_interface(identity, nic_id):
    nics = app.systems.get_nics(identity)

    for nic in nics:
        if nic['id'] == nic_id:
            return flask.render_template(
                'ethernet_interface.json', identity=identity, nic=nic)

    raise error.NotFound()


@app.route('/redfish/v1/Systems/<identity>/Processors',
           methods=['GET'])
@api_utils.ensure_instance_access
@api_utils.returns_json
def processors_collection(identity):
    processors = app.systems.get_processors(identity)

    return flask.render_template(
        'processors_collection.json', identity=identity,
        processors=processors)


@app.route('/redfish/v1/Systems/<identity>/Processors/<processor_id>',
           methods=['GET'])
@api_utils.ensure_instance_access
@api_utils.returns_json
def processor(identity, processor_id):
    processors = app.systems.get_processors(identity)

    for proc in processors:
        if proc['id'] == processor_id:
            return flask.render_template(
                'processor.json', identity=identity, processor=proc)

    raise error.NotFound()


@app.route('/redfish/v1/Systems/<identity>/Actions/ComputerSystem.Reset',
           methods=['POST'])
@api_utils.ensure_instance_access
@api_utils.returns_json
def system_reset_action(identity):
    reset_type = flask.request.json.get('ResetType')

    app.systems.set_power_state(identity, reset_type)

    app.logger.info('System "%s" power state set to "%s"',
                    identity, reset_type)

    return '', 204


@app.route('/redfish/v1/Systems/<identity>/BIOS', methods=['GET'])
@api_utils.ensure_instance_access
@api_utils.returns_json
def bios(identity):
    bios = app.systems.get_bios(identity)

    app.logger.debug('Serving BIOS for system "%s"', identity)

    return flask.render_template(
        'bios.json',
        identity=identity,
        bios_current_attributes=json.dumps(bios, sort_keys=True, indent=6))


@app.route('/redfish/v1/Systems/<identity>/BIOS/Settings',
           methods=['GET', 'PATCH'])
@api_utils.ensure_instance_access
@api_utils.returns_json
def bios_settings(identity):

    if flask.request.method == 'GET':
        bios = app.systems.get_bios(identity)

        app.logger.debug('Serving BIOS Settings for system "%s"', identity)

        return flask.render_template(
            'bios_settings.json',
            identity=identity,
            bios_pending_attributes=json.dumps(bios, sort_keys=True, indent=6))

    elif flask.request.method == 'PATCH':
        attributes = flask.request.json.get('Attributes')

        app.systems.set_bios(identity, attributes)

        app.logger.info('System "%s" BIOS attributes "%s" updated',
                        identity, attributes)
        return '', 204


@app.route('/redfish/v1/Systems/<identity>/BIOS/Actions/Bios.ResetBios',
           methods=['POST'])
@api_utils.ensure_instance_access
@api_utils.returns_json
def system_reset_bios(identity):
    app.systems.reset_bios(identity)

    app.logger.info('BIOS for system "%s" reset', identity)

    return '', 204


@app.route('/redfish/v1/Systems/<identity>/SimpleStorage',
           methods=['GET'])
@api_utils.ensure_instance_access
@api_utils.returns_json
def simple_storage_collection(identity):
    simple_storage_controllers = (
        app.systems.get_simple_storage_collection(identity))

    return flask.render_template(
        'simple_storage_collection.json', identity=identity,
        simple_storage_controllers=simple_storage_controllers)


@app.route('/redfish/v1/Systems/<identity>/SimpleStorage/<simple_storage_id>',
           methods=['GET'])
@api_utils.ensure_instance_access
@api_utils.returns_json
def simple_storage(identity, simple_storage_id):
    simple_storage_controllers = (
        app.systems.get_simple_storage_collection(identity))
    try:
        storage_controller = simple_storage_controllers[simple_storage_id]
    except KeyError:
        app.logger.debug('"%s" Simple Storage resource was not found')
        raise error.NotFound()
    return flask.render_template('simple_storage.json', identity=identity,
                                 simple_storage=storage_controller)


@app.route('/redfish/v1/Systems/<identity>/Storage',
           methods=['GET'])
@api_utils.ensure_instance_access
@api_utils.returns_json
def storage_collection(identity):
    uuid = app.systems.uuid(identity)

    storage_col = app.storage.get_storage_col(uuid)

    return flask.render_template(
        'storage_collection.json', identity=identity,
        storage_col=storage_col)


@app.route('/redfish/v1/Systems/<identity>/Storage/<storage_id>',
           methods=['GET'])
@api_utils.ensure_instance_access
@api_utils.returns_json
def storage(identity, storage_id):
    uuid = app.systems.uuid(identity)
    storage_col = app.storage.get_storage_col(uuid)

    for stg in storage_col:
        if stg['Id'] == storage_id:
            return flask.render_template(
                'storage.json', identity=identity, storage=stg)

    raise error.NotFound()


@app.route('/redfish/v1/Systems/<identity>/Storage/<stg_id>/Drives/<drv_id>',
           methods=['GET'])
@api_utils.ensure_instance_access
@api_utils.returns_json
def drive_resource(identity, stg_id, drv_id):
    uuid = app.systems.uuid(identity)
    drives = app.drives.get_drives(uuid, stg_id)

    for drv in drives:
        if drv['Id'] == drv_id:
            return flask.render_template(
                'drive.json', identity=identity, storage_id=stg_id, drive=drv)

    raise error.NotFound()


@app.route('/redfish/v1/Systems/<identity>/Storage/<storage_id>/Volumes',
           methods=['GET', 'POST'])
@api_utils.ensure_instance_access
@api_utils.returns_json
def volumes_collection(identity, storage_id):
    uuid = app.systems.uuid(identity)

    if flask.request.method == 'GET':

        vol_col = app.volumes.get_volumes_col(uuid, storage_id)

        vol_ids = []
        for vol in vol_col:
            vol_id = app.systems.find_or_create_storage_volume(vol)
            if not vol_id:
                app.volumes.delete_volume(uuid, storage_id, vol)
            else:
                vol_ids.append(vol_id)

        return flask.render_template(
            'volume_collection.json', identity=identity,
            storage_id=storage_id, volume_col=vol_ids)

    elif flask.request.method == 'POST':
        data = {
            "Name": flask.request.json.get('Name'),
            "VolumeType": flask.request.json.get('VolumeType'),
            "CapacityBytes": flask.request.json.get('CapacityBytes'),
            "Id": str(os.getpid()) + datetime.now().strftime("%H%M%S")
        }
        data['libvirtVolName'] = data['Id']
        new_id = app.systems.find_or_create_storage_volume(data)
        if new_id:
            app.volumes.add_volume(uuid, storage_id, data)
            app.logger.debug('New storage volume created with ID "%s"',
                             new_id)
            vol_url = ("/redfish/v1/Systems/%s/Storage/%s/"
                       "Volumes/%s" % (identity, storage_id, new_id))
            return flask.Response(status=201,
                                  headers={'Location': vol_url})


@app.route('/redfish/v1/Systems/<identity>/Storage/<stg_id>/Volumes/<vol_id>',
           methods=['GET'])
@api_utils.ensure_instance_access
@api_utils.returns_json
def volume(identity, stg_id, vol_id):
    uuid = app.systems.uuid(identity)
    vol_col = app.volumes.get_volumes_col(uuid, stg_id)

    for vol in vol_col:
        if vol['Id'] == vol_id:
            vol_id = app.systems.find_or_create_storage_volume(vol)
            if not vol_id:
                app.volumes.delete_volume(uuid, stg_id, vol)
            else:
                return flask.render_template(
                    'volume.json', identity=identity, storage_id=stg_id,
                    volume=vol)

    raise error.NotFound()


@app.route('/redfish/v1/Registries')
@api_utils.returns_json
def registry_file_collection():
    app.logger.debug('Serving registry file collection')

    return flask.render_template(
        'registry_file_collection.json')


@app.route('/redfish/v1/Registries/BiosAttributeRegistry.v1_0_0')
@api_utils.returns_json
def bios_attribute_registry_file():
    app.logger.debug('Serving BIOS attribute registry file')

    return flask.render_template(
        'bios_attribute_registry_file.json')


@app.route('/redfish/v1/Registries/Messages')
@api_utils.returns_json
def message_registry_file():
    app.logger.debug('Serving message registry file')

    return flask.render_template(
        'message_registry_file.json')


@app.route('/redfish/v1/Systems/Bios/BiosRegistry')
@api_utils.returns_json
def bios_registry():
    app.logger.debug('Serving BIOS registry')

    return flask.render_template('bios_registry.json')


@app.route('/redfish/v1/Registries/Messages/Registry')
@api_utils.returns_json
def message_registry():
    app.logger.debug('Serving message registry')

    return flask.render_template('message_registry.json')


def parse_args():
    parser = argparse.ArgumentParser('sushy-emulator')
    parser.add_argument('--config',
                        type=str,
                        help='Config file path. Can also be set via '
                             'environment variable SUSHY_EMULATOR_CONFIG.')
    parser.add_argument('--debug', action='store_true',
                        help='Enables debug mode when running sushy-emulator.')
    parser.add_argument('-i', '--interface',
                        type=str,
                        help='IP address of the local interface to listen '
                             'at. Can also be set via config variable '
                             'SUSHY_EMULATOR_LISTEN_IP. Default is all '
                             'local interfaces.')
    parser.add_argument('-p', '--port',
                        type=int,
                        help='TCP port to bind the server to.  Can also be '
                             'set via config variable '
                             'SUSHY_EMULATOR_LISTEN_PORT. Default is 8000.')
    parser.add_argument('--ssl-certificate',
                        type=str,
                        help='SSL certificate to use for HTTPS. Can also be '
                        'set via config variable SUSHY_EMULATOR_SSL_CERT.')
    parser.add_argument('--ssl-key',
                        type=str,
                        help='SSL key to use for HTTPS. Can also be set'
                        'via config variable SUSHY_EMULATOR_SSL_KEY.')
    backend_group = parser.add_mutually_exclusive_group()
    backend_group.add_argument('--os-cloud',
                               type=str,
                               help='OpenStack cloud name. Can also be set '
                                    'via environment variable OS_CLOUD or '
                                    'config variable SUSHY_EMULATOR_OS_CLOUD.'
                               )
    backend_group.add_argument('--libvirt-uri',
                               type=str,
                               help='The libvirt URI. Can also be set via '
                                    'environment variable '
                                    'SUSHY_EMULATOR_LIBVIRT_URI. '
                                    'Default is qemu:///system')
    backend_group.add_argument('--fake', action='store_true',
                               help='Use the fake driver. Can also be set '
                                    'via environmnet variable '
                                    'SUSHY_EMULATOR_FAKE_DRIVER.')

    return parser.parse_args()


def main():

    args = parse_args()

    app.debug = args.debug

    app.configure(config_file=args.config)

    if args.os_cloud:
        app.config['SUSHY_EMULATOR_OS_CLOUD'] = args.os_cloud

    if args.libvirt_uri:
        app.config['SUSHY_EMULATOR_LIBVIRT_URI'] = args.libvirt_uri

    if args.fake:
        app.config['SUSHY_EMULATOR_FAKE_DRIVER'] = True

    else:
        for envvar in ('SUSHY_EMULATOR_LIBVIRT_URL',  # backward compatibility
                       'SUSHY_EMULATOR_LIBVIRT_URI'):
            envvar = os.environ.get(envvar)
            if envvar:
                app.config['SUSHY_EMULATOR_LIBVIRT_URI'] = envvar

    if args.interface:
        app.config['SUSHY_EMULATOR_LISTEN_IP'] = args.interface

    if args.port:
        app.config['SUSHY_EMULATOR_LISTEN_PORT'] = args.port

    if args.ssl_certificate:
        app.config['SUSHY_EMULATOR_SSL_CERT'] = args.ssl_certificate

    if args.ssl_key:
        app.config['SUSHY_EMULATOR_SSL_KEY'] = args.ssl_key

    ssl_context = None
    ssl_certificate = app.config.get('SUSHY_EMULATOR_SSL_CERT')
    ssl_key = app.config.get('SUSHY_EMULATOR_SSL_KEY')

    if ssl_certificate and ssl_key:
        ssl_context = ssl.SSLContext(ssl.PROTOCOL_TLSv1_2)
        ssl_context.load_cert_chain(ssl_certificate, ssl_key)

    app.run(host=app.config.get('SUSHY_EMULATOR_LISTEN_IP'),
            port=app.config.get('SUSHY_EMULATOR_LISTEN_PORT', 8000),
            ssl_context=ssl_context)

    return 0


if __name__ == '__main__':
    sys.exit(main())
